from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, Sequence, TypeVar, List, Any

from sklearn.exceptions import NotFittedError # type: ignore
import numpy as np
import numpy.typing as npt
DT = TypeVar("DT")
CT = TypeVar("CT")
LT = TypeVar("LT")

class BaseVectorizer(ABC, Generic[DT]):
    """This is the :class:`~abc.ABC` specifies a generic vectorizer.
    Vectorizers transform raw data examples into feature vectors.
    Given a data type `DT`, it specifies the methods :meth:`~.fit`
    that initializes or fits the vectorizer. The method :meth:`~.transform`
    transforms the data into vector form.
    """    
    _name = "BaseVectorizer"
    
    def __init__(self):
        self._fitted = False
    
    @property
    def fitted(self) -> bool:
        """Check if the vectorizer has been fitted

        Returns
        -------
        bool
            True if the vectorizer has been fitted
        """        
        return self._fitted

    @abstractmethod
    def fit(self, x_data: Sequence[DT], **kwargs: Any) -> BaseVectorizer[DT]:
        """Fit the vectorizer according to the data in the given 
        :class:`~collections.abc.Sequence`.

        Parameters
        ----------
        x_data : Sequence[DT]
            A Sequence of examples with type `DT`.

        Returns
        -------
        BaseVectorizer[DT]
            A fitted vectorizer for data with type `DT`

        Examples
        --------
        Assume the creation of a vectorizer and a sequence of data examples
        in the variable `data_list`
        
        >>> vectorizer = BaseVectorizer[DT]()
        >>> vectorizer = vectorizer.fit(data_list)
        """        
        pass

    @abstractmethod
    def transform(self, x_data: Sequence[DT], **kwargs: Any) -> npt.NDArray[Any]: # type: ignore
        """Transform a list raw data points to a feature matrix 
        according to the fitted vectorizer

        Parameters
        ----------
        x_data : Sequence[DT]
            A sequence of raw data examples with length `n_examples`

        Returns
        -------
        npt.NDArray[Any]
            A feature matrix with shape `(n_examples, n_features)`

        Examples
        --------
        Assume the vectorizer is fitted
        
        >>> x_mat = vectorizer.transform(x_data)
        """        
        pass

    @abstractmethod
    def fit_transform(self, x_data: Sequence[DT], **kwargs: Any) -> npt.NDArray[Any]: # type: ignore
        """Transform a list of data to a feature matrix. The transformation
        is based on the data contained in the parameter `x_data`. Subsequent
        transformations with :meth:`~.transform()` will be based on the fit
        of the data provided in this call.

        Parameters
        ----------
        x_data : Sequence[DT]
            A sequence of raw data examples with length `n_examples`

        Returns
        -------
        npt.NDArray[Any]
            A feature matrix with shape `(n_examples, n_features)`

        Examples
        --------
        Assume the vectorizer is fitted
        
        >>> x_mat = vectorizer.fit_transform(x_data)
        """
        pass

class SeparateContextVectorizer(ABC, Generic[DT, CT]):
    """This :class:`~abc.ABC` specifies a generic vectorizer for data types
    that consists of two parts that have to be fitted or configured according
    to different specifications. The feature vectors of the two parts
    are concatenated for each example.

    The two parts are referred to the `data` part and the `context` part.
    This vectorizer contains two inner vectorizer, one for the data part and 
    one for the context part respectively.

    Arguments
    ---------
    data_vectorizer : BaseVectorizer[DT]
            The vectorizer for the data part
    context_vectorizer : BaseVectorizer[CT]
            The vectorizer for the context part
    
    Examples
    --------
    Construction:

    >>> from sklearn.feature_extraction.text import TfidfVectorizer
    >>> data_vectorizer = SklearnVectorizer[str](TfidfVectorizer())
    >>> context_vectorizer = Doc2VecVectorizer[str]()
    >>> vectorizer = SeparateContextVectorizer[str, str](data_vectorizer, 
    ...     context_vectorizer)

    Fitting:

    >>> x_data = ["This...", "Another text...", ... "Last Text"]
    >>> x_context_data = ["Surrounding text", ... , "Another text"]
    >>> vectorizer = vectorizer.fit(x_data, x_context_data)

    Transforming:

    >>> x_mat = vectorizer.transform(x_data, x_context_data)
    """    
    
    _name = "SeparateContextVectorizer"
    
    def __init__(
            self,
            data_vectorizer: BaseVectorizer[DT],
            context_vectorizer: BaseVectorizer[CT]
        ):
        self.data_vectorizer = data_vectorizer
        self.context_vectorizer = context_vectorizer

    @property
    def fitted(self) -> bool:
        """Check if the vectorizer has been fitted

        Returns
        -------
        bool
            True if the vectorizer has been fitted
        """        
        return self.data_vectorizer.fitted and self.context_vectorizer.fitted

    def fit(
            self,
            x_data: Sequence[DT],
            context_data: Sequence[CT],
            **kwargs: Any) -> SeparateContextVectorizer[DT, CT]:
        """Fit the vectorizer according to the data in the given 
        :class:`~collections.abc.Sequence` s.

        Parameters
        ----------
        x_data : Sequence[DT]
            The data parts
        context_data : Sequence[CT]
            The contexts parts

        Returns
        -------
        SeparateContextVectorizer[DT, CT]
            A fitted vectorizer

        Examples
        --------
        Fitting this vectorizer can be performed as follows:

        >>> x_data = ["This...", "Another text...", ... "Last Text"]
        >>> x_context_data = ["Surrounding text", ... , "Another text"]
        >>> data_vectorizer = SklearnVectorizer[str](TfidfVectorizer())
        >>> context_vectorizer = Doc2VecVectorizer[str]()
        >>> vectorizer = SeparateContextVectorizer[str, str](
        ...     data_vectorizer, 
        ...     context_vectorizer)
        >>> vectorizer = vectorizer.fit(x_data, x_context_data)

        Warning
        -------
        We assume that the variables `x_data` and `context_data` are sequences
        of equal length.
        """        
        self.data_vectorizer.fit(x_data, **kwargs)
        self.context_vectorizer.fit(context_data, **kwargs)
        return self

    def transform(
            self,
            x_data: Sequence[DT],
            context_data: Sequence[CT],
            **kwargs: Any) -> npt.NDArray[Any]: # type: ignore
        """Transform a list raw data points to a feature matrix 
        according to the fitted vectorizers


        Parameters
        ----------
        x_data : Sequence[DT]
            A sequence with data parts of the data points of length `n_docs`
        context_data : Sequence[CT]
            A sequence with context part of the data points of length `n_docs`

        Returns
        -------
        npt.NDArray[Any]
            A feature matrix of concatenated vectors with shape
            `(n_docs, n_features_data + n_features_context)`

        Raises
        ------
        NotFittedError
            If the model is not fitted
        
        Warning
        -------
        We assume that the variables `x_data` and `context_data` are sequences
        of equal length and that the indices of the sequences correspond to the
        same data point.
        """        
        if self.fitted:
            data_part: npt.NDArray[Any] = self.data_vectorizer.transform(x_data, **kwargs) # type: ignore
            context_part: npt.NDArray[Any] = self.context_vectorizer.transform( # type: ignore
                context_data, **kwargs) # type: ignore
            return np.concatenate((data_part, context_part), axis=1) # type: ignore
        raise NotFittedError

    def fit_transform(
            self,
            x_data: Sequence[DT],
            context_data: Sequence[CT],
            **kwargs: Any) -> npt.NDArray[Any]: # type: ignore
        """Fit and transform a list raw data points to a feature matrix 
        according to the fitted vectorizers. Subsequent
        transformations with :meth:`~.transform()` will be based on the fit
        of the data provided in this call.


        Parameters
        ----------
        x_data : Sequence[DT]
            A sequence with data parts of the data points of length `n_docs`
        context_data : Sequence[CT]
            A sequence with context part of the data points of length `n_docs`

        Returns
        -------
        npt.NDArray[Any]
            A feature matrix of concatenated vectors with shape
            `(n_docs, n_features_data + n_features_context)`
        """        
        self.fit(x_data, **kwargs)
        return self.transform(x_data, context_data, **kwargs) # type: ignore


class StackVectorizer(BaseVectorizer[DT], Generic[DT]):
    """This :class:`~abc.ABC` specifies a generic vectorizer that consists of 
    several vectorizers that are fitted on the same data points. 
    
    The feature vectors of the contained vectorizers are concatenated in the 
    transform step, according to the order they are specified in the 
    constructor (argument order).
    
    Arguments
    ----------
    vectorizer : BaseVectorizer[DT]
        At least one vectorizer is required
    *vectorizers: BaseVectorizer[DT]
        Any number of vectorizers for the same data type

    Examples
    --------
    Construction

    >>> tf_idf = SklearnVectorizer[str](TfidfVectorizer())
    >>> doc2vec = Doc2VecVectorizer[str]()
    >>> count = SklearnVectorizer[str](CountVectorizer())
    >>> vectorizer = StackVectorizer[str](tfidf, doc2vec, count)

    Fitting

    >>> x_data = ["This...", "Another text...", ... "Last Text"]
    >>> vectorizer = vectorizer.fit(x_data)

    Transforming

    >>> another_data = ["Another test text", ... , "Another text"]
    >>> x_mat = vectorizer.transform(x_data)    
    """    
    
    vectorizers: List[BaseVectorizer[DT]]
    """The internal vectorizers are stored in this list"""    
    
    _name = "StackVectorizer"

    def __init__(self,
                 vectorizer: BaseVectorizer[DT],
                 *vectorizers: BaseVectorizer[DT]) -> None:
        """[summary]

        Parameters
        ----------
        vectorizer : BaseVectorizer[DT]
            [description]
        """        
        super().__init__()
        self.vectorizers = [vectorizer, *vectorizers]

    def fit(self, x_data: Sequence[DT], **kwargs: Any) -> StackVectorizer[DT]:
        for vec in self.vectorizers:
            vec.fit(x_data, **kwargs)
        return self

    def fitted(self) -> bool:
        return all([vec.fitted for vec in self.vectorizers])

    def transform(self, x_data: Sequence[DT], **kwargs: Any) -> npt.NDArray[Any]: # type: ignore
        if self.fitted:
            sub_vectors = [ # type: ignore
                vec.transform(x_data, **kwargs) # type: ignore
                for vec in self.vectorizers]
            return np.concatenate(sub_vectors, axis=1) # type: ignore
        raise NotFittedError

    def fit_transform(self, x_data: Sequence[DT], **kwargs: Any) -> npt.NDArray[Any]: # type: ignore
        self.fit(x_data, **kwargs)
        return self.transform(x_data, **kwargs) # type: ignore



from __future__ import annotations
from abc import ABC, abstractmethod

import itertools
import logging
import random
from typing import (
    Any,
    Callable,
    Deque,
    Dict,
    FrozenSet,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)
from typing_extensions import Self, TypeVar
from uuid import uuid4

import numpy as np  # type: ignore
import pandas as pd  # type: ignore

from ..environment import AbstractEnvironment
from ..activelearning.base import ActiveLearner
from ..activelearning.estimator import Estimator

from ..typehints import KT, LT, IT, DT, VT, RT

LOGGER = logging.getLogger(__name__)

AT = TypeVar("AT", bound="ActiveLearner[Any, Any, Any, Any, Any, Any]", covariant=True)


class Initializer(ABC, Generic[IT, KT, LT]):
    @abstractmethod
    def __call__(
        self, learner: ActiveLearner[IT, KT, DT, VT, RT, LT]
    ) -> ActiveLearner[IT, KT, DT, VT, RT, LT]:
        raise NotImplementedError

    @classmethod
    def builder(cls, *args, **kwargs) -> Callable[..., Self]:
        def builder_func(*args, **kwargs) -> Self:
            return cls()

        return builder_func


class IdentityInitializer(Initializer[IT, KT, LT], Generic[IT, KT, LT]):
    def __call__(
        self, learner: ActiveLearner[IT, KT, DT, VT, RT, LT]
    ) -> ActiveLearner[IT, KT, DT, VT, RT, LT]:
        return learner


class RandomInitializer(Initializer[IT, KT, LT], Generic[IT, KT, LT]):
    def __init__(self, sample_size: int = 1) -> None:
        self.sample_size = sample_size

    def get_random_sample_for_label(
        self, learner: ActiveLearner[IT, KT, Any, Any, Any, LT], label: LT
    ) -> Sequence[KT]:
        docs = random.sample(
            learner.env.truth.get_instances_by_label(label), self.sample_size
        )
        return docs

    def get_initialization_sample(
        self, learner: ActiveLearner[IT, KT, Any, Any, Any, LT]
    ) -> Sequence[KT]:
        docs = list(
            itertools.chain.from_iterable(
                map(
                    lambda lbl: self.get_random_sample_for_label(learner, lbl),
                    learner.env.labels.labelset,
                )
            )
        )
        return docs

    def add_doc(
        self, learner: ActiveLearner[IT, KT, Any, Any, Any, LT], identifier: KT
    ):
        doc = learner.env.dataset[identifier]
        labels = learner.env.truth.get_labels(doc)
        learner.env.labels.set_labels(doc, *labels)
        learner.set_as_labeled(doc)

    def __call__(
        self, learner: ActiveLearner[IT, KT, DT, VT, RT, LT]
    ) -> ActiveLearner[IT, KT, DT, VT, RT, LT]:
        docs = self.get_initialization_sample(learner)
        for doc in docs:
            self.add_doc(learner, doc)
        return learner

    @classmethod
    def builder(cls, sample_size: int, *args, **kwargs) -> Callable[..., Self]:
        def builder_func(*args, **kwargs) -> Self:
            return cls(sample_size)

        return builder_func


class UniformInitializer(RandomInitializer[IT, KT, LT], Generic[IT, KT, LT]):
    def __call__(
        self, learner: ActiveLearner[IT, KT, DT, VT, RT, LT]
    ) -> ActiveLearner[IT, KT, DT, VT, RT, LT]:
        if not isinstance(learner, Estimator):
            return super().__call__(learner)
        docs = self.get_initialization_sample(learner)
        for sublearner in learner.learners:
            for doc in docs:
                self.add_doc(sublearner, doc)
                self.add_doc(learner, doc)
        return learner


class SeparateInitializer(RandomInitializer[IT, KT, LT], Generic[IT, KT, LT]):
    def __call__(
        self, learner: ActiveLearner[IT, KT, DT, VT, RT, LT]
    ) -> ActiveLearner[IT, KT, DT, VT, RT, LT]:
        if not isinstance(learner, Estimator):
            return super().__call__(learner)
        for sublearner in learner.learners:
            docs = self.get_initialization_sample(learner)
            for doc in docs:
                self.add_doc(sublearner, doc)
                self.add_doc(learner, doc)
        return learner


class PositiveUniformInitializer(RandomInitializer[IT, KT, LT], Generic[IT, KT, LT]):
    def __init__(self, pos_label: LT, neg_label: LT, sample_size: int = 1) -> None:
        super().__init__(sample_size)
        self.pos_label = pos_label
        self.neg_label = neg_label

    def __call__(
        self, learner: ActiveLearner[IT, KT, DT, VT, RT, LT]
    ) -> ActiveLearner[IT, KT, DT, VT, RT, LT]:
        if not isinstance(learner, Estimator):
            return super().__call__(learner)
        pos_docs = self.get_random_sample_for_label(learner, self.pos_label)
        for sublearner in learner.learners:
            for doc in pos_docs:
                self.add_doc(sublearner, doc)
                self.add_doc(learner, doc)
        for sublearner in learner.learners:
            neg_docs = self.get_random_sample_for_label(learner, self.neg_label)
            for doc in neg_docs:
                self.add_doc(sublearner, doc)
                self.add_doc(learner, doc)
        return learner

    @classmethod
    def builder(cls, sample_size: int, *args, **kwargs) -> Callable[..., Self]:
        def builder_func(pos_label: LT, neg_label: LT, *args, **kwargs) -> Self:
            return cls(pos_label, neg_label, sample_size)

        return builder_func



from enum import Enum


class ALCatalog:
    class Paradigm(Enum):
        POOLBASED = "Poolbased"
        ESTIMATOR = "Estimator"
        CYCLE_ESTIMATOR = "CycleEstimator"
        ENSEMBLE = "Ensemble"
        RETRY_ESTIMATOR = "NewEstimator"
        PROBABILITY_BASED = "ProbabilityBased"
        LABEL_PROBABILITY_BASED = "LabelProbabilityBased"
        PROBABILITY_BASED_ENSEMBLE = "ProbabilityBasedEnsemble"
        LABEL_PROBABILITY_BASED_ENSEMBLE = "LabelProbabilityBasedEnsemble"
        LABEL_MIN_PROB_ENSEMBLE = "LabelMinCountProbEnsemble"
        CUSTOM = "CUSTOM"

    class QueryType(str, Enum):
        RANDOM_SAMPLING = "RandomSampling"
        LEAST_CONFIDENCE = "LeastConfidence"
        NEAR_DECISION_BOUNDARY = "NearDecisionBoundary"
        MARGIN_SAMPLING = "MarginSampling"
        MOST_CERTAIN = "MostCertain"
        MAX_ENTROPY = "MaxEntropy"
        INTERLEAVE = "InterleaveAL"
        LABELMAXIMIZER = "LabelMaximizer"
        LABELUNCERTAINTY = "LabelUncertainty"
        MIN_ENTROPY = "MinEntropy"
        MOST_CONFIDENCE = "MostConfidence"
        PRELABELED = "Prelabeled"
        LABELMAXIMIZER_NEW = "LabelMaximizerNew"
        LABELUNCERTAINTY_NEW = "LabelUncertaintyNew"
        RANDOM_ML = "RandomML"

    class CustomMethods(Enum):
        AUTOTAR = "AUTOTAR"
        AUTOSTOP = "AUTOSTOP"



from __future__ import annotations

from typing import Sequence, Optional, Callable, List, Dict, Any
from tempfile import NamedTemporaryFile

import numpy as np
import numpy.typing as npt
from sklearn.exceptions import NotFittedError # type: ignore

from gensim.models.doc2vec import Doc2Vec, TaggedLineDocument # type: ignore
from gensim.utils import save_as_line_sentence # type: ignore

from .base import BaseVectorizer
from ..utils import SaveableInnerModel


def get_line_docs(documents: Sequence[str]) -> TaggedLineDocument:
    file = NamedTemporaryFile(delete=False, mode="w+b")
    file.close()
    corpus = [[element] for element in documents]
    save_as_line_sentence(corpus, file.name)
    tld = TaggedLineDocument(file.name)
    return tld


def split_tokenizer(text: str) -> List[str]:
    return text.split(' ')


DocTokenizer = Callable[..., List[str]]


class Doc2VecVectorizer(BaseVectorizer[str], SaveableInnerModel[Doc2Vec]):
    _name = "Doc2Vec"
    innermodel: Optional[Doc2Vec]
    tokenizer: DocTokenizer

    def __init__(
        self,
        d2v_params: Dict[str, Any],
        tokenizer: DocTokenizer = split_tokenizer,
        storage_location: Optional[str]=None,
        filename: Optional[str] = None
    ) -> None:
        BaseVectorizer.__init__(self) # type: ignore
        self.tokenizer = tokenizer # type: ignore
        self.d2v_params = d2v_params
        self.innermodel = None
        SaveableInnerModel.__init__(self, self.innermodel, storage_location, filename)

    def fit(self, x_data: Sequence[str], **kwargs: Any) -> Doc2VecVectorizer:
        self.innermodel = Doc2Vec(
            documents=get_line_docs(x_data), **self.d2v_params)
        self.innermodel.delete_temporary_training_data( # type: ignore
            keep_doctags_vectors=True, keep_inference=True) 
        self._fitted = True
        return self

    @SaveableInnerModel.load_model_fallback
    def transform(self, x_data: Sequence[str], **kwargs: Any) -> npt.NDArray[Any]: # type: ignore
        if self.fitted and self.innermodel is not None:
            return np.array( # type: ignore
                [
                    self.innermodel.infer_vector(self.tokenizer(doc)) # type: ignore
                    for doc in x_data
                ]) 
        raise NotFittedError

    def fit_transform(self, x_data: Sequence[str], **kwargs: Any) -> npt.NDArray[Any]: # type: ignore
        self.fit(x_data)
        return self.transform(x_data) # type: ignore

    def save(self) -> None:
        if self.innermodel is not None:
            self.innermodel.save(self.filepath) # type: ignore
            self.saved = True

    def load(self) -> None:
        self.innermodel = Doc2Vec.load(self.filepath) # type: ignore


from __future__ import annotations

import os

from typing import Generic, TypeVar, Optional, List, Callable, Any

import functools
import pickle
import uuid

MT = TypeVar("MT")

F = TypeVar('F', bound=Callable[..., Any])

class SaveableInnerModel(Generic[MT]):
    _name = "SaveableInnerModel"

    def __init__(self, innermodel: Optional[MT], storage_location: Optional[str], filename: Optional[str] = None, taboo_fields: Optional[List[str]] = None):
        self.storage_location = storage_location
        self.saved = False
        self.innermodel = innermodel
        if filename is None:
            self.filename = self._generate_random_file_name()
        else:
            self.filename = filename
        self.taboo_fields = {"innermodel": None}
        if taboo_fields is not None:
            for field in taboo_fields:
                self.taboo_fields[field] = None

    def _generate_random_file_name(self) -> str:
        """Generates a random filename

        Returns
        -------
        str
            A random file name
        """
        gen_uuid = uuid.uuid4()
        filename = f"classifier_{self._name}_{gen_uuid}.data"
        return filename

    @property
    def filepath(self) -> Optional[str]:
        if self.storage_location is not None:
            return os.path.join(self.storage_location, self.filename)
        return None

    @property
    def has_storage_location(self) -> bool:
        return self.storage_location is not None

    @property
    def is_stored(self) -> bool:
        return self.saved

    @staticmethod
    def load_model_fallback(func: F) -> F:
        @functools.wraps(func)
        def wrapper(self: SaveableInnerModel[MT], *args: Any, **kwargs: Any) -> Any:
            if not self.is_loaded and self.is_stored:
                self.load()
            return func(self, *args, **kwargs)
        return wrapper # type: ignore

    @property
    def is_loaded(self) -> bool:
        return self.innermodel is not None

    def __getstate__(self):
        if not self.has_storage_location:
            return self.__dict__
        self.save()
        state = {key: value for (key, value) in self.__dict__.items(
        ) if key not in self.taboo_fields}
        state = {**state.copy(), **self.taboo_fields}
        return state

    def save(self) -> None:
        assert self.filepath is not None
        with open(self.filepath, mode="wb") as filehandle:
            pickle.dump(self.innermodel, filehandle)
        self.saved = True

    def load(self) -> None:
        assert self.filepath is not None
        with open(self.filepath, mode="rb") as filehandle:
            self.innermodel = pickle.load(filehandle)



from .base import ActiveLearner
from .factory import ActiveLearningFactory
from .poolbased import PoolBasedAL
from .random import RandomSampling
from .uncertainty import EntropySampling, LeastConfidence
from .catalog import ALCatalog



# Copyright (C) 2021 The ALLib Authors. All Rights Reserved.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from typing import Any, TypeVar

from instancelib import Instance

KT = TypeVar("KT") # The Key Type of the Dataset (mostly int for Primary Keys, but strings are also possible)
DT = TypeVar("DT") # The Type of the data point
VT = TypeVar("VT") # The Type of the vector
RT = TypeVar("RT") # The Type of the representation of the data point
LT = TypeVar("LT") # The Type of the labels
CT = TypeVar("CT") # The Type of the context
LVT = TypeVar("LVT") # The Type of the encoded label vector
PVT = TypeVar("PVT") # The Type of the probability vector
IT = TypeVar("IT", bound="Instance[Any, Any, Any, Any]", covariant=True)



from typing import Any, Dict
from ..module import ModuleCatalog as Cat


def add_identifier(config: Dict[str, Any], identifier: str) -> Dict[str, Any]:
    id_dict = {"identifier": identifier}
    new_dict = {**config, **id_dict}
    return new_dict


al_config_svm = {
    "paradigm": Cat.AL.Paradigm.LABEL_PROBABILITY_BASED,
    "query_type": Cat.AL.QueryType.LABELMAXIMIZER,
    "label": "Relevant",
    "machinelearning": {
        "sklearn_model": Cat.ML.SklearnModel.SVC,
        "model_configuration": {
            "kernel": "linear",
            "probability": True,
            "class_weight": "balanced",
        },
        "task": Cat.ML.Task.BINARY,
        "balancer": {"type": Cat.BL.Type.DOUBLE, "config": {}},
    },
}

al_config_lr = {
    "paradigm": Cat.AL.Paradigm.LABEL_PROBABILITY_BASED,
    "query_type": Cat.AL.QueryType.LABELMAXIMIZER,
    "label": "Relevant",
    "machinelearning": {
        "sklearn_model": Cat.ML.SklearnModel.LOGISTIC,
        "model_configuration": {
            "solver": "lbfgs",
            "C": 1.0,
            "max_iter": 10000,
        },
        "task": Cat.ML.Task.BINARY,
        "balancer": {"type": Cat.BL.Type.DOUBLE, "config": {}},
    },
}

al_config_nb = {
    "paradigm": Cat.AL.Paradigm.LABEL_PROBABILITY_BASED,
    "query_type": Cat.AL.QueryType.LABELMAXIMIZER,
    "label": "Relevant",
    "machinelearning": {
        "sklearn_model": Cat.ML.SklearnModel.NAIVE_BAYES,
        "model_configuration": {
            "alpha": 3.822,
        },
        "task": Cat.ML.Task.BINARY,
        "balancer": {"type": Cat.BL.Type.DOUBLE, "config": {}},
    },
}
al_config_lgbm = {
    "paradigm": Cat.AL.Paradigm.LABEL_PROBABILITY_BASED,
    "query_type": Cat.AL.QueryType.LABELMAXIMIZER,
    "label": "Relevant",
    "machinelearning": {
        "sklearn_model": Cat.ML.SklearnModel.LGBM,
        "model_configuration": {},
        "task": Cat.ML.Task.BINARY,
        "balancer": {"type": Cat.BL.Type.DOUBLE, "config": {}},
    },
}
al_config_svm_random = {
    "paradigm": Cat.AL.Paradigm.PROBABILITY_BASED_ENSEMBLE,
    "strategies": [
        {"query_type": Cat.AL.QueryType.MAX_ENTROPY},
        {"query_type": Cat.AL.QueryType.MOST_CONFIDENCE},
    ],
    "machinelearning": {
        "sklearn_model": Cat.ML.SklearnModel.SVC,
        "model_configuration": {
            "kernel": "linear",
            "probability": True,
            "class_weight": "balanced",
        },
        "task": Cat.ML.Task.MULTILABEL,
        "balancer": {"type": Cat.BL.Type.IDENTITY, "config": {}},
        "mc_method": Cat.ML.MulticlassMethod.ONE_VS_REST,
    },
}

al_config_ensemble_prob = {
    "paradigm": Cat.AL.Paradigm.PROBABILITY_BASED_ENSEMBLE,
    "strategies": [
        {"query_type": Cat.AL.QueryType.MAX_ENTROPY},
        {"query_type": Cat.AL.QueryType.MOST_CONFIDENCE},
    ],
    "machinelearning": {
        "sklearn_model": Cat.ML.SklearnModel.SVC,
        "model_configuration": {
            "kernel": "linear",
            "probability": True,
            "class_weight": "balanced",
        },
        "task": Cat.ML.Task.MULTILABEL,
        "balancer": {"type": Cat.BL.Type.IDENTITY, "config": {}},
        "mc_method": Cat.ML.MulticlassMethod.ONE_VS_REST,
    },
}

al_config_ensemble_labelprob = {
    "paradigm": Cat.AL.Paradigm.LABEL_PROBABILITY_BASED_ENSEMBLE,
    "strategy": Cat.AL.QueryType.LABELUNCERTAINTY_NEW,
    "machinelearning": {
        "sklearn_model": Cat.ML.SklearnModel.SVC,
        "model_configuration": {
            "kernel": "linear",
            "probability": True,
            "class_weight": "balanced",
        },
        "task": Cat.ML.Task.MULTILABEL,
        "balancer": {"type": Cat.BL.Type.IDENTITY, "config": {}},
        "mc_method": Cat.ML.MulticlassMethod.ONE_VS_REST,
    },
}
al_config_ensemble_random = {
    "paradigm": Cat.AL.Paradigm.LABEL_PROBABILITY_BASED_ENSEMBLE,
    "strategy": Cat.AL.QueryType.RANDOM_ML,
    "machinelearning": {
        "sklearn_model": Cat.ML.SklearnModel.SVC,
        "model_configuration": {
            "kernel": "linear",
            "probability": True,
            "class_weight": "balanced",
        },
        "task": Cat.ML.Task.MULTILABEL,
        "balancer": {"type": Cat.BL.Type.IDENTITY, "config": {}},
        "mc_method": Cat.ML.MulticlassMethod.ONE_VS_REST,
    },
}


al_config_rf = {
    "paradigm": Cat.AL.Paradigm.LABEL_PROBABILITY_BASED,
    "query_type": Cat.AL.QueryType.LABELMAXIMIZER,
    "label": "Relevant",
    "machinelearning": {
        "sklearn_model": Cat.ML.SklearnModel.RANDOM_FOREST,
        "model_configuration": {
            "n_estimators": 100,
            "max_features": 10,
        },
        "task": Cat.ML.Task.BINARY,
        "balancer": {"type": Cat.BL.Type.DOUBLE, "config": {}},
    },
}
al_config_unc = {
    "paradigm": Cat.AL.Paradigm.LABEL_PROBABILITY_BASED,
    "query_type": Cat.AL.QueryType.LABELUNCERTAINTY,
    "machinelearning": {
        "sklearn_model": Cat.ML.SklearnModel.NAIVE_BAYES,
        "model_configuration": {
            "alpha": 3.822,
        },
        "task": Cat.ML.Task.BINARY,
        "balancer": {"type": Cat.BL.Type.DOUBLE, "config": {}},
    },
}
al_config_svm_multilabel = {
    "paradigm": Cat.AL.Paradigm.PROBABILITY_BASED,
    "query_type": Cat.AL.QueryType.MARGIN_SAMPLING,
    "machinelearning": {
        "sklearn_model": Cat.ML.SklearnModel.SVC,
        "model_configuration": {
            "kernel": "linear",
            "probability": True,
            "class_weight": "balanced",
        },
        "task": Cat.ML.Task.MULTILABEL,
        "balancer": {"type": Cat.BL.Type.IDENTITY, "config": {}},
        "mc_method": Cat.ML.MulticlassMethod.ONE_VS_REST,
    },
}

al_config_random = {
    "paradigm": Cat.AL.Paradigm.POOLBASED,
    "query_type": Cat.AL.QueryType.RANDOM_SAMPLING,
}
mixed_estimator = {
    "paradigm": Cat.AL.Paradigm.ESTIMATOR,
    "learners": [
        al_config_nb,
        al_config_svm,
        al_config_rf,
        al_config_lgbm,
    ],
}

naive_bayes_estimator = {
    "paradigm": Cat.AL.Paradigm.ESTIMATOR,
    "learners": [
        add_identifier(al_config_nb, "NaiveBayes1"),
        add_identifier(al_config_nb, "NaiveBayes2"),
        add_identifier(al_config_nb, "NaiveBayes3"),
        add_identifier(al_config_nb, "NaiveBayes4"),
    ],
}
svm_estimator = {
    "paradigm": Cat.AL.Paradigm.ESTIMATOR,
    "learners": [
        add_identifier(al_config_svm, "SVM1"),
        add_identifier(al_config_svm, "SVM2"),
        add_identifier(al_config_svm, "SVM3"),
        add_identifier(al_config_svm, "SVM4"),
    ],
}

rasch_estimator = {
    "paradigm": Cat.AL.Paradigm.ESTIMATOR,
    "learners": [
        add_identifier(al_config_nb, "NaiveBayes"),
        add_identifier(al_config_svm, "SVM"),
        add_identifier(al_config_rf, "RandomForest"),
    ],
}

rasch_nblrrf = {
    "paradigm": Cat.AL.Paradigm.ESTIMATOR,
    "learners": [
        add_identifier(al_config_nb, "NaiveBayes"),
        add_identifier(al_config_rf, "RandomForest"),
        add_identifier(al_config_lr, "LogisticRegression"),
    ],
}
rasch_nblrrfsvm = {
    "paradigm": Cat.AL.Paradigm.ESTIMATOR,
    "learners": [
        add_identifier(al_config_nb, "NaiveBayes"),
        add_identifier(al_config_rf, "RandomForest"),
        add_identifier(al_config_lr, "LogisticRegression"),
        add_identifier(al_config_svm, "SVM"),
    ],
}
rasch_nblrrflgbm = {
    "paradigm": Cat.AL.Paradigm.ESTIMATOR,
    "learners": [
        add_identifier(al_config_nb, "NaiveBayes"),
        add_identifier(al_config_rf, "RandomForest"),
        add_identifier(al_config_lr, "LogisticRegression"),
        add_identifier(al_config_lgbm, "LGBM"),
    ],
}
rasch_nblrrflgbmrand = {
    "paradigm": Cat.AL.Paradigm.ESTIMATOR,
    "learners": [
        add_identifier(al_config_nb, "NaiveBayes"),
        add_identifier(al_config_rf, "RandomForest"),
        add_identifier(al_config_lr, "LogisticRegression"),
        add_identifier(al_config_lgbm, "LGBM"),
        add_identifier(al_config_random, "Random"),
    ],
}

rasch_lr = {
    "paradigm": Cat.AL.Paradigm.ESTIMATOR,
    "learners": [
        add_identifier(al_config_nb, "NaiveBayes"),
        add_identifier(al_config_svm, "SVM"),
        add_identifier(al_config_lr, "LogisticRegression"),
    ],
}
rasch_rf = {
    "paradigm": Cat.AL.Paradigm.ESTIMATOR,
    "learners": [
        add_identifier(al_config_rf, "RandomForest1"),
        add_identifier(al_config_rf, "RandomForest2"),
        add_identifier(al_config_rf, "RandomForest3"),
    ],
}

rasch_random_estimator = {
    "paradigm": Cat.AL.Paradigm.ESTIMATOR,
    "learners": [
        add_identifier(al_config_nb, "NaiveBayes"),
        add_identifier(al_config_svm, "SVM"),
        add_identifier(al_config_rf, "RandomForest"),
        add_identifier(al_config_random, "Random3"),
    ],
}

# rasch_random_estimator = {
#     "paradigm": Cat.AL.Paradigm.ESTIMATOR,
#     "learners": [
#         add_identifier(al_config_random, "Random1"),
#         add_identifier(al_config_random, "Random2"),
#         add_identifier(al_config_random, "Random3"),
#     ]
# }


al_config_est3 = {
    "paradigm": Cat.AL.Paradigm.ESTIMATOR,
    "learners": [
        add_identifier(al_config_nb, "NaiveBayes1"),
        add_identifier(al_config_svm, "SVM"),
        add_identifier(al_config_rf, "RandomForest"),
        add_identifier(al_config_nb, "NaiveBayes4"),
    ],
}
al_config_est4 = {
    "paradigm": Cat.AL.Paradigm.ESTIMATOR,
    "learners": [
        add_identifier(al_config_nb, "NaiveBayes1"),
        add_identifier(al_config_svm, "SVM"),
        add_identifier(al_config_rf, "RandomForest"),
        add_identifier(al_config_lr, "LogisticRegression"),
    ],
}
al_config_est5 = {
    "paradigm": Cat.AL.Paradigm.ESTIMATOR,
    "learners": [
        add_identifier(al_config_nb, "NaiveBayes1"),
        add_identifier(al_config_svm, "SVM"),
        add_identifier(al_config_rf, "RandomForest"),
        add_identifier(al_config_svm, "SVM2"),
        add_identifier(al_config_svm, "SVM3"),
    ],
}

al_config_ens = {
    "paradigm": Cat.AL.Paradigm.ENSEMBLE,
    "learners": [
        al_config_nb,
        al_config_svm,
        al_config_rf,
        al_config_unc,
    ],
    "probabilities": [0.35, 0.35, 0.2, 0.1],
    "machinelearning": {
        "sklearn_model": Cat.ML.SklearnModel.NAIVE_BAYES,
        "model_configuration": {},
        "task": Cat.ML.Task.BINARY,
        "balancer": {"type": Cat.BL.Type.DOUBLE, "config": {}},
    },
}

al_config_entropy = {
    "paradigm": Cat.AL.Paradigm.PROBABILITY_BASED,
    "query_type": Cat.AL.QueryType.MOST_CONFIDENCE,
    "machinelearning": {
        "sklearn_model": Cat.ML.SklearnModel.SVC,
        "model_configuration": {
            "kernel": "linear",
            "probability": True,
            "class_weight": "balanced",
        },
        "task": Cat.ML.Task.MULTILABEL,
        "balancer": {"type": Cat.BL.Type.DOUBLE, "config": {}},
        "mc_method": Cat.ML.MulticlassMethod.ONE_VS_REST,
    },
}


env_config = {"environment_type": Cat.ENV.Type.MEMORY}

tf_idf5000 = {
    "datatype": Cat.FE.DataType.TEXTINSTANCE,
    "vec_type": Cat.FE.VectorizerType.STACK,
    "vectorizers": [
        {
            "vec_type": Cat.FE.VectorizerType.SKLEARN,
            "sklearn_vec_type": Cat.FE.SklearnVecType.TFIDF_VECTORIZER,
            "sklearn_config": {"max_features": 5000},
        }
    ],
}

tf_idf_autotar = {
    "datatype": Cat.FE.DataType.TEXTINSTANCE,
    "vec_type": Cat.FE.VectorizerType.STACK,
    "vectorizers": [
        {
            "vec_type": Cat.FE.VectorizerType.SKLEARN,
            "sklearn_vec_type": Cat.FE.SklearnVecType.TFIDF_VECTORIZER,
            "sklearn_config": {
                "stop_words": "english",
                "min_df": 2,
                "max_features": 3000,
            },
        }
    ],
}

autotar = {
    "paradigm": Cat.AL.Paradigm.CUSTOM,
    "query_type": Cat.AL.QueryType.MOST_CONFIDENCE,
    "machinelearning": {
        "sklearn_model": Cat.ML.SklearnModel.SVC,
        "model_configuration": {
            "kernel": "linear",
            "probability": True,
            "class_weight": "balanced",
        },
        "task": Cat.ML.Task.MULTILABEL,
        "balancer": {"type": Cat.BL.Type.DOUBLE, "config": {}},
        "mc_method": Cat.ML.MulticlassMethod.ONE_VS_REST,
    },
}



from re import M
from typing import Any, Generic, Tuple

import numpy as np
import numpy.typing as npt

from ..activelearning import ActiveLearner
from ..activelearning.autostop import AutoStopLearner
from ..typehints import DT, IT, KT, LT, RT, VT
from .base import AbstractEstimator, Estimate

from numba import njit


def calc_fo_mask(learner: AutoStopLearner[Any, Any, Any, Any, Any, Any], it: int) -> npt.NDArray[Any]:
    labeled = learner.cumulative_sampled[it]
    fo_mask = np.array([int(k in labeled) for k in learner.key_seq]).reshape((1,-1))
    return fo_mask

def calc_so_mask(fo_mask: npt.NDArray[Any], big_n: int) -> npt.NDArray[Any]:
    mat = np.tile(fo_mask, (big_n, 1))
    mat_t = mat.T
    so_mask = mat * mat_t
    return so_mask

def horvitz_thompson_point(learner: AutoStopLearner[Any, Any, Any, Any, Any, Any], 
                           it: int) -> float:
    
    fo_mask = calc_fo_mask(learner, it)
    fo_prob = learner.fo_inclusion_probabilities(it)
    ys = learner.label_vector[it]
    estimate = np.sum(fo_mask * (ys / fo_prob))
    return estimate

def horvitz_thompson_var1(
        learner: AutoStopLearner[Any, Any, Any, Any, Any, Any],
        it: int) -> float:
    fo_mask = calc_fo_mask(learner, it)    
    big_n = len(learner.env.dataset)
    so_mask = calc_so_mask(fo_mask, big_n)
    
    ys = learner.label_vector[it]
    fo_pi = learner.fo_inclusion_probabilities(it)
    so_pij = learner.so_inclusion_probabilities(it)
    
    part1 = 1.0 / fo_pi ** 2 - 1.0 / fo_pi
    yi_2 = ys ** 2
    
    # 1/(pi_i*pi_j) - 1/pi_ij
    M = np.tile(fo_pi, (big_n, 1))
    MT = M.T
    part2 = 1.0 / (M * MT) - 1.0 / so_pij
    np.fill_diagonal(part2, 0.0)  # set diagonal values to zero, because summing part2 do not include diagonal values

    #  y_i * y_j
    M = np.tile(ys, (big_n, 1))
    MT = M.T
    yi_yj = M*MT
    
    variance = np.sum(fo_mask * part1*yi_2) + np.sum(so_mask *part2*yi_yj)
    return variance

def horvitz_thompson_var2(learner: AutoStopLearner[Any, Any, Any, Any, Any, Any], it: int) -> float:
    fo_mask = calc_fo_mask(learner, it)
    big_n = len(learner.env.dataset)
    v = len(learner.cumulative_sampled[it])
    ys = learner.label_vector[it]
    total = horvitz_thompson_point(learner, it)
    fo_prob = learner.fo_inclusion_probabilities(it)

    # (v * y_i / pi_i - total)**2
    variance2 = (v * ys / fo_prob - total) ** 2
    variance2 = (big_n - v) / big_n / v / (v - 1) * np.sum(fo_mask*variance2)
    return variance2



class HorvitzThompsonVar1(AbstractEstimator[IT, KT, DT, VT, RT, LT], 
                      Generic[IT, KT, DT, VT, RT, LT]):
    def __init__(self) -> None:
        super().__init__()
        self.it = -1
        self.estimate = Estimate(float("nan"), float("nan"), float("nan"))
    

    def calculate(self, learner: AutoStopLearner, it: int) -> Estimate:
        point = horvitz_thompson_point(learner, it)
        variance = horvitz_thompson_var1(learner, it)
        error = np.sqrt(variance)
        estimate = Estimate(point, point - error, point + error)
        return estimate

    def __call__(self, learner: ActiveLearner[Any, KT, DT, VT, RT, LT], label: LT) -> Estimate:
        assert isinstance(learner, AutoStopLearner)
        if learner.cumulative_fo:
            it = min(max(learner.cumulative_fo.keys()), learner.it)
            if it > 1 and self.it != it:
                self.estimate = self.calculate(learner, it)
        return self.estimate
        
class HorvitzThompsonVar2(HorvitzThompsonVar1[IT, KT, DT, VT, RT, LT], 
                      Generic[IT, KT, DT, VT, RT, LT]):
    def calculate(self, learner: AutoStopLearner, it: int) -> Estimate:
        point = horvitz_thompson_point(learner, it)
        variance = horvitz_thompson_var2(learner, it)
        error = np.sqrt(variance)
        estimate = Estimate(point, point - error, point + error)
        return estimate

    

class HorvitzThompsonLoose(HorvitzThompsonVar1[IT, KT, DT, VT, RT, LT], 
                      Generic[IT, KT, DT, VT, RT, LT]):
    def calculate(self, learner: AutoStopLearner, it: int) -> Estimate:
        point = horvitz_thompson_point(learner, it)
        estimate = Estimate(point, point, point)
        return estimate



import array as arr
import copy as cp
import logging
import numpy as np
import numpy.typing as npt
import sys
from scipy import optimize
import scipy.sparse.csc as csc
from scipy import sparse
import scipy
import warnings
#warnings.simplefilter('error')

class QN_S3VM:
    """
    L-BFGS optimizer for semi-supervised support vector machines (S3VM).
    """
    def __init__(self, X_l, L_l, X_u, random_generator = None, ** kw):
        """
        Initializes the model. Detects automatically if dense or sparse data is provided.
        Keyword arguments:
        X_l -- patterns of labeled part of the data
        L_l -- labels of labeled part of the data
        X_u -- patterns of unlabeled part of the data
        random_generator -- particular instance of a random_generator (default None)
        kw -- additional parameters for the optimizer
        lam -- regularization parameter lambda (default 1, must be a float > 0)
        lamU -- cost parameter that determines influence of unlabeled patterns (default 1, must be float > 0)
        sigma -- kernel width for RBF kernel (default 1.0, must be a float > 0)
        kernel_type -- "Linear" or "RBF" (default "Linear")
        numR -- implementation of subset of regressors. If None is provided, all patterns are used
                (no approximation). Must fulfill 0 <= numR <= len(X_l) + len(X_u) (default None)
        estimate_r -- desired ratio for positive and negative assigments for 
                      unlabeled patterns (-1.0 <= estimate_r <= 1.0). If estimate_r=None, 
                      then L_l is used to estimate this ratio (in case len(L_l) >= 
                      minimum_labeled_patterns_for_estimate_r. Otherwise use estimate_r = 0.0
                      (default None)
        minimum_labeled_patterns_for_estimate_r -- see above (default 0)
        BFGS_m -- BFGS parameter (default 50)
        BFGS_maxfun -- BFGS parameter, maximum number of function calls (default 500)
        BFGS_factr -- BFGS parameter (default 1E12)
        BFGS_pgtol -- BFGS parameter (default 1.0000000000000001e-05)
        """
        self.__model = None
        # Initiate model for sparse data
        if isinstance(X_l, csc.csc_matrix):
            self.__data_type = "sparse"
            self.__model = QN_S3VM_Sparse(X_l, L_l, X_u, random_generator, ** kw)
        # Initiate model for dense data
        elif (isinstance(X_l[0], list)) or (isinstance(X_l[0], npt.NDArray[Any])):
            self.__data_type = "dense"
            self.__model = QN_S3VM_Dense(X_l, L_l, X_u, random_generator, ** kw)
        # Data format unknown
        if self.__model == None:
            logging.info("Data format for patterns is unknown.")
            sys.exit(0)

    def train(self):
        """
        Training phase.
        Returns:
        The computed partition for the unlabeled patterns.
        """
        return self.__model.train()

    def getPredictions(self, X, real_valued=False):
        """
        Computes the predicted labels for a given set of patterns
        Keyword arguments:
        X -- The set of patterns 
        real_valued -- If True, then the real prediction values are returned
        Returns:
        The predictions for the list X of patterns.
        """
        return self.__model.getPredictions(X, real_valued=real_valued)

    def predict(self, x):
        """
        Predicts a label (-1 or +1) for the pattern
        Keyword arguments:
        x -- The pattern 
        Returns:
        The prediction for x.
        """
        return self.__model.predict(x)

    def predictValue(self, x):
        """
        Computes f(x) for a given pattern (see Representer Theorem)
    
        Keyword arguments:
        x -- The pattern 
        Returns:
        The (real) prediction value for x.
        """
        return self.__model.predictValue(x)

    def getNeededFunctionCalls(self):
        """
        Returns the number of function calls needed during 
        the optimization process.
        """
        return self.__model.getNeededFunctionCalls()

    def mygetPreds(self, X, real_valued=False):
        return self.__model.mygetPreds(X, real_valued)

############################################################################################
############################################################################################
class QN_S3VM_Dense:

    """
    BFGS optimizer for semi-supervised support vector machines (S3VM).
    Dense Data
    """
    parameters = {
    'lam': 1,
    'lamU':1,
    'sigma': 1,
    'kernel_type': "Linear",
    'numR':None,
    'estimate_r':None,
    'minimum_labeled_patterns_for_estimate_r':0,
    'BFGS_m':50,
    'BFGS_maxfun':500,
    'BFGS_factr':1E12,
    'BFGS_pgtol':1.0000000000000001e-05,
    'BFGS_verbose':-1,
    'surrogate_s':3.0,
    'surrogate_gamma':20.0,
    'breakpoint_for_exp':500
    }

    def __init__(self, X_l, L_l, X_u, random_generator, ** kw):
        """
        Intializes the S3VM optimizer.
        """
        self.__random_generator = random_generator
        self.__X_l, self.__X_u, self.__L_l = X_l, X_u, L_l
        assert len(X_l) == len(L_l)
        self.__X = cp.deepcopy(self.__X_l)
        self.__X.extend(cp.deepcopy(self.__X_u))
        self.__size_l, self.__size_u, self.__size_n = len(X_l), len(X_u), len(X_l) + len(X_u)
        self.__matrices_initialized = False
        self.__setParameters( ** kw)
        self.__kw = kw

    def train(self):
        """
        Training phase.
        Returns:
        The computed partition for the unlabeled patterns.
        """
        indi_opt = self.__optimize()
        self.__recomputeModel(indi_opt)
        predictions = self.__getTrainingPredictions(self.__X)
        return predictions

    def mygetPreds(self, X, real_valued=False):
        KNR = self.__kernel.computeKernelMatrix(X, self.__Xreg)
        KNU_bar = self.__kernel.computeKernelMatrix(X, self.__X_u_subset, symmetric=False)
        KNU_bar_horizontal_sum = (1.0 / len(self.__X_u_subset)) * KNU_bar.sum(axis=1)
        KNR = KNR - KNU_bar_horizontal_sum - self.__KU_barR_vertical_sum + self.__KU_barU_bar_sum
        preds = KNR * self.__c[0:self.__dim-1,:] + self.__c[self.__dim-1,:]
        return preds

    def getPredictions(self, X, real_valued=False):
        """
        Computes the predicted labels for a given set of patterns
        Keyword arguments:
        X -- The set of patterns 
        real_valued -- If True, then the real prediction values are returned
        Returns:
        The predictions for the list X of patterns.
        """
        KNR = self.__kernel.computeKernelMatrix(X, self.__Xreg)
        KNU_bar = self.__kernel.computeKernelMatrix(X, self.__X_u_subset, symmetric=False)
        KNU_bar_horizontal_sum = (1.0 / len(self.__X_u_subset)) * KNU_bar.sum(axis=1)
        KNR = KNR - KNU_bar_horizontal_sum - self.__KU_barR_vertical_sum + self.__KU_barU_bar_sum
        preds = KNR * self.__c[0:self.__dim-1,:] + self.__c[self.__dim-1,:]
        if real_valued == True:
            return preds.flatten().tolist()[0]
        else:
            return np.sign(np.sign(preds)+0.1).flatten().tolist()[0]
    
    def predict(self, x):
        """
        Predicts a label for the pattern
        Keyword arguments:
        x -- The pattern 
        Returns:
        The prediction for x.
        """
        return self.getPredictions([x], real_valued=False)[0]
        
    def predictValue(self, x):
        """
        Computes f(x) for a given pattern (see Representer Theorem)
    
        Keyword arguments:
        x -- The pattern 
        Returns:
        The (real) prediction value for x.
        """
        return self.getPredictions([x], real_valued=True)[0]

    def getNeededFunctionCalls(self):
        """
        Returns the number of function calls needed during 
        the optimization process.
        """
        return self.__needed_function_calls

    def __setParameters(self,  ** kw):
        for attr, val in kw.items():
            self.parameters[attr] = val
        self.__lam = float(self.parameters['lam'])
        assert self.__lam > 0
        self.__lamU = float(self.parameters['lamU'])
        assert self.__lamU > 0
        self.__lam_Uvec = [float(self.__lamU)*i for i in [0,0.000001,0.0001,0.01,0.1,0.5,1]]
        self.__sigma = float(self.parameters['sigma'])
        assert self.__sigma > 0
        self.__kernel_type = str(self.parameters['kernel_type'])
        if self.parameters['numR'] != None:
            self.__numR = int(self.parameters['numR'])
            assert (self.__numR <= len(self.__X)) and (self.__numR > 0)
        else:
            self.__numR = len(self.__X)
        self.__regressors_indices = sorted(self.__random_generator.sample( range(0,len(self.__X)), self.__numR ))
        self.__dim = self.__numR + 1 # add bias term b
        self.__minimum_labeled_patterns_for_estimate_r = float(self.parameters['minimum_labeled_patterns_for_estimate_r'])
        # If reliable estimate is available or can be estimated, use it, otherwise
        # assume classes to be balanced (i.e., estimate_r=0.0)
        if self.parameters['estimate_r'] != None:
            self.__estimate_r = float(self.parameters['estimate_r'])
        elif len(self.__L_l) >= self.__minimum_labeled_patterns_for_estimate_r:
            self.__estimate_r = (1.0 / len(self.__L_l)) * np.sum(self.__L_l)
        else:
            self.__estimate_r = 0.0
        self.__BFGS_m = int(self.parameters['BFGS_m'])
        self.__BFGS_maxfun = int(self.parameters['BFGS_maxfun'])
        self.__BFGS_factr = float(self.parameters['BFGS_factr'])
        # This is a hack for 64 bit systems (Linux). The machine precision 
        # is different for the BFGS optimizer (Fortran code) and we fix this by:
        is_64bits = sys.maxsize > 2**32
        if is_64bits:
            logging.debug("64-bit system detected, modifying BFGS_factr!")
            self.__BFGS_factr = 0.000488288*self.__BFGS_factr
        self.__BFGS_pgtol = float(self.parameters['BFGS_pgtol'])
        self.__BFGS_verbose = int(self.parameters['BFGS_verbose'])
        self.__surrogate_gamma = float(self.parameters['surrogate_gamma'])
        self.__s = float(self.parameters['surrogate_s'])
        self.__breakpoint_for_exp = float(self.parameters['breakpoint_for_exp'])
        self.__b = self.__estimate_r
        # size of unlabeled patterns to estimate mean (used for balancing constraint)
        self.__max_unlabeled_subset_size = 1000


    def __optimize(self):
        logging.debug("Starting optimization with BFGS ...")
        self.__needed_function_calls = 0
        self.__initializeMatrices()
        # starting point
        c_current = np.zeros(self.__dim, np.float64)
        c_current[self.__dim-1] = self.__b
        # Annealing sequence.
        for i in range(len(self.__lam_Uvec)):
            self.__lamU = self.__lam_Uvec[i]
            # crop one dimension (in case the offset b is fixed)
            c_current = c_current[:self.__dim-1]
            c_current = self.__localSearch(c_current)
            # reappend it if needed
            c_current = np.append(c_current, self.__b)
        f_opt = self.__getFitness(c_current)
        return c_current, f_opt

    def __localSearch(self, start):
        c_opt, f_opt, d = optimize.fmin_l_bfgs_b(self.__getFitness, start, m=self.__BFGS_m, \
                            fprime=self.__getFitness_Prime, maxfun=self.__BFGS_maxfun, factr=self.__BFGS_factr,\
                            pgtol=self.__BFGS_pgtol, iprint=self.__BFGS_verbose)
        self.__needed_function_calls += int(d['funcalls'])
        return c_opt

    def __initializeMatrices(self):
        if self.__matrices_initialized == False:
            logging.debug("Initializing matrices...")
            # Initialize labels
            x = arr.array('i')
            for l in self.__L_l:
                x.append(l)
            self.__YL = np.mat(x, dtype=np.float64)
            self.__YL = self.__YL.transpose()
            # Initialize kernel matrices
            if (self.__kernel_type == "Linear"):
                self.__kernel = LinearKernel()
            elif (self.__kernel_type == "RBF"):
                self.__kernel = RBFKernel(self.__sigma)
            self.__Xreg = (np.mat(self.__X)[self.__regressors_indices,:].tolist())
            self.__KLR = self.__kernel.computeKernelMatrix(self.__X_l,self.__Xreg, symmetric=False)
            self.__KUR = self.__kernel.computeKernelMatrix(self.__X_u,self.__Xreg, symmetric=False)
            self.__KNR = cp.deepcopy(np.bmat([[self.__KLR], [self.__KUR]]))
            self.__KRR = self.__KNR[self.__regressors_indices,:]
            # Center patterns in feature space (with respect to approximated mean of unlabeled patterns in the feature space)
            subset_unlabled_indices = sorted(self.__random_generator.sample( range(0,len(self.__X_u)), min(self.__max_unlabeled_subset_size, len(self.__X_u)) ))
            self.__X_u_subset = (np.mat(self.__X_u)[subset_unlabled_indices,:].tolist())
            self.__KNU_bar = self.__kernel.computeKernelMatrix(self.__X, self.__X_u_subset, symmetric=False)
            self.__KNU_bar_horizontal_sum = (1.0 / len(self.__X_u_subset)) * self.__KNU_bar.sum(axis=1)
            self.__KU_barR = self.__kernel.computeKernelMatrix(self.__X_u_subset, self.__Xreg, symmetric=False)
            self.__KU_barR_vertical_sum = (1.0 / len(self.__X_u_subset)) * self.__KU_barR.sum(axis=0)
            self.__KU_barU_bar = self.__kernel.computeKernelMatrix(self.__X_u_subset, self.__X_u_subset, symmetric=False)
            self.__KU_barU_bar_sum = (1.0 / (len(self.__X_u_subset)))**2 * self.__KU_barU_bar.sum()
            self.__KNR = self.__KNR - self.__KNU_bar_horizontal_sum - self.__KU_barR_vertical_sum + self.__KU_barU_bar_sum
            self.__KRR = self.__KNR[self.__regressors_indices,:]
            self.__KLR = self.__KNR[range(0,len(self.__X_l)),:]
            self.__KUR = self.__KNR[range(len(self.__X_l),len(self.__X)),:]
            self.__matrices_initialized = True

    def __getFitness(self,c):
        # Check whether the function is called from the bfgs solver 
        # (that does not optimize the offset b) or not
        if len(c) == self.__dim - 1:
            c = np.append(c, self.__b)
        c = np.mat(c)
        b = c[:,self.__dim-1].T
        c_new = c[:,0:self.__dim-1].T
        preds_labeled = self.__surrogate_gamma*(1.0 - np.multiply(self.__YL, self.__KLR * c_new + b))
        preds_unlabeled = self.__KUR * c_new + b
        # This vector has a "one" for each "numerically instable" entry; "zeros" for "good ones". 
        preds_labeled_conflict_indicator = np.sign(np.sign(preds_labeled/self.__breakpoint_for_exp - 1.0) + 1.0)
        # This vector has a one for each good entry and zero otherwise
        preds_labeled_good_indicator = (-1)*(preds_labeled_conflict_indicator - 1.0)
        preds_labeled_for_conflicts = np.multiply(preds_labeled_conflict_indicator,preds_labeled) 
        preds_labeled = np.multiply(preds_labeled,preds_labeled_good_indicator)
        # Compute values for good entries
        preds_labeled_log_exp = np.log(1.0 + np.exp(preds_labeled))
        # Compute values for instable entries
        preds_labeled_log_exp = np.multiply(preds_labeled_good_indicator, preds_labeled_log_exp)
        # Replace critical values with values 
        preds_labeled_final = preds_labeled_log_exp + preds_labeled_for_conflicts
        term1 = (1.0/(self.__surrogate_gamma*self.__size_l)) * np.sum(preds_labeled_final)
        preds_unlabeled_squared = np.multiply(preds_unlabeled,preds_unlabeled)
        term2 = (float(self.__lamU)/float(self.__size_u))*np.sum(np.exp(-self.__s * preds_unlabeled_squared))
        term3 = self.__lam * (c_new.T * self.__KRR * c_new)
        return (term1 + term2 + term3)[0,0]

    def __getFitness_Prime(self,c):
        # Check whether the function is called from the bfgs solver 
        # (that does not optimize the offset b) or not
        if len(c) == self.__dim - 1:
            c = np.append(c, self.__b)
        c = np.mat(c)
        b = c[:,self.__dim-1].T
        c_new = c[:,0:self.__dim-1].T
        preds_labeled = self.__surrogate_gamma * (1.0 - np.multiply(self.__YL, self.__KLR * c_new + b))
        preds_unlabeled = (self.__KUR * c_new + b)
        # This vector has a "one" for each "numerically instable" entry; "zeros" for "good ones". 
        preds_labeled_conflict_indicator = np.sign(np.sign(preds_labeled/self.__breakpoint_for_exp - 1.0) + 1.0)
        # This vector has a one for each good entry and zero otherwise
        preds_labeled_good_indicator = (-1)*(preds_labeled_conflict_indicator - 1.0)
        preds_labeled = np.multiply(preds_labeled,preds_labeled_good_indicator)
        preds_labeled_exp = np.exp(preds_labeled)
        term1 = np.multiply(preds_labeled_exp, 1.0/(1.0 + preds_labeled_exp))
        term1 = np.multiply(preds_labeled_good_indicator, term1)
        # Replace critical values with "1.0"
        term1 = term1 + preds_labeled_conflict_indicator
        term1 = np.multiply(self.__YL, term1)
        preds_unlabeled_squared_exp_f = np.multiply(preds_unlabeled,preds_unlabeled)
        preds_unlabeled_squared_exp_f = np.exp(-self.__s * preds_unlabeled_squared_exp_f)
        preds_unlabeled_squared_exp_f = np.multiply(preds_unlabeled_squared_exp_f, preds_unlabeled)
        term1 = (-1.0/self.__size_l) * (term1.T * self.__KLR).T
        term2 = ((-2.0 * self.__s * self.__lamU)/float(self.__size_u)) * (preds_unlabeled_squared_exp_f.T * self.__KUR).T
        term3 = 2*self.__lam*(self.__KRR * c_new)
        return np.array((term1 + term2 + term3).T)[0]

    def __recomputeModel(self, indi):
        self.__c = np.mat(indi[0]).T

    def __getTrainingPredictions(self, X, real_valued=False):
        preds = self.__KNR * self.__c[0:self.__dim-1,:] + self.__c[self.__dim-1,:]
        if real_valued == True:
            return preds.flatten().tolist()[0]
        else:
            # import ipdb
            # ipdb.set_trace()
            preds = np.asarray(preds)
            return np.sign(np.sign(preds)+0.1).flatten().tolist()[0]

    def __check_matrix(self, M):
        smallesteval = scipy.linalg.eigvalsh(M, eigvals=(0,0))[0]
        if smallesteval < 0.0:
            shift = abs(smallesteval) + 0.0000001 
            M = M + shift
        return M

############################################################################################
############################################################################################
class QN_S3VM_Sparse:
    """
    BFGS optimizer for semi-supervised support vector machines (S3VM).
    Sparse Data
    """
    parameters = {
    'lam': 1,
    'lamU':1,
    'estimate_r':None,
    'minimum_labeled_patterns_for_estimate_r':0,
    'BFGS_m':50,
    'BFGS_maxfun':500,
    'BFGS_factr':1E12,
    'BFGS_pgtol':1.0000000000000001e-05,
    'BFGS_verbose':-1,
    'surrogate_s':3.0,
    'surrogate_gamma':20.0,
    'breakpoint_for_exp':500
    }


    def __init__(self, X_l, L_l, X_u, random_generator, ** kw):
        """
        Intializes the S3VM optimizer.
        """
        self.__random_generator = random_generator
        # This is a nuisance, but we may need to pad extra dimensions to either X_l or X_u
        # in case the highest feature indices appear only in one of the two data matrices
        if X_l.shape[1] > X_u.shape[1]:
            X_u = sparse.hstack([X_u, sparse.coo_matrix(X_u.shape[0], X_l.shape[1] - X_u.shape[1])])
        elif X_l.shape[1] < X_u.shape[1]:
            X_l = sparse.hstack([X_l, sparse.coo_matrix(X_l.shape[0], X_u.shape[1] - X_u.shape[1])])
        # We vertically stack the data matrices into one big matrix
        X = sparse.vstack([X_l, X_u])
        self.__size_l, self.__size_u, self.__size_n = X_l.shape[0], X_u.shape[0], X_l.shape[0]+ X_u.shape[0]
        x = arr.array('i')
        for l in L_l:
            x.append(int(l))
        self.__YL = np.mat(x, dtype=np.float64)
        self.__YL = self.__YL.transpose()
        self.__setParameters( ** kw)
        self.__kw = kw
        self.X_l = X_l.tocsr()
        self.X_u = X_u.tocsr()
        self.X = X.tocsr()
        # compute mean of unlabeled patterns
        self.__mean_u = self.X_u.mean(axis=0)
        self.X_u_T = X_u.tocsc().T
        self.X_l_T = X_l.tocsc().T
        self.X_T = X.tocsc().T

    def train(self):
        """
        Training phase.
        Returns:
        The computed partition for the unlabeled patterns.
        """
        indi_opt = self.__optimize()
        self.__recomputeModel(indi_opt)
        predictions = self.getPredictions(self.X)
        return predictions

    def getPredictions(self, X, real_valued=False):
        """
        Computes the predicted labels for a given set of patterns
        Keyword arguments:
        X -- The set of patterns 
        real_valued -- If True, then the real prediction values are returned
        Returns:
        The predictions for the list X of patterns.
        """
        c_new = self.__c[:self.__dim-1]
        W = self.X.T*c_new - self.__mean_u.T*np.sum(c_new)
        # Again, possibility of dimension mismatch due to use of sparse matrices
        if X.shape[1] > W.shape[0]:
            X = X[:,range(W.shape[0])]
        if X.shape[1] < W.shape[0]:
            W = W[range(X.shape[1])]
        X = X.tocsc()
        preds = X * W + self.__b
        if real_valued == True:
            return preds.flatten().tolist()[0]
        else:
            return np.sign(np.sign(preds)+0.1).flatten().tolist()[0]

    def predict(self, x):
        """
        Predicts a label for the pattern
        Keyword arguments:
        x -- The pattern 
        Returns:
        The prediction for x.
        """
        return self.getPredictions([x], real_valued=False)[0]
        
    def predictValue(self, x):
        """
        Computes f(x) for a given pattern (see Representer Theorem)
    
        Keyword arguments:
        x -- The pattern 
        Returns:
        The (real) prediction value for x.
        """
        return self.getPredictions([x], real_valued=True)[0]

    def getNeededFunctionCalls(self):
        """
        Returns the number of function calls needed during 
        the optimization process.
        """
        return self.__needed_function_calls

    def __setParameters(self,  ** kw):
        for attr, val in kw.items():
            self.parameters[attr] = val
        self.__lam = float(self.parameters['lam'])
        assert self.__lam > 0
        self.__lamU = float(self.parameters['lamU'])
        assert self.__lamU > 0
        self.__lam_Uvec = [float(self.__lamU)*i for i in [0,0.000001,0.0001,0.01,0.1,0.5,1]]
        self.__minimum_labeled_patterns_for_estimate_r = float(self.parameters['minimum_labeled_patterns_for_estimate_r'])
        # If reliable estimate is available or can be estimated, use it, otherwise
        # assume classes to be balanced (i.e., estimate_r=0.0)
        if self.parameters['estimate_r'] != None:
            self.__estimate_r = float(self.parameters['estimate_r'])
        elif self.__YL.shape[0] > self.__minimum_labeled_patterns_for_estimate_r:
            self.__estimate_r = (1.0 / self.__YL.shape[0]) * np.sum(self.__YL[0:])
        else:
            self.__estimate_r = 0.0
        self.__dim = self.__size_n + 1 # for offset term b
        self.__BFGS_m = int(self.parameters['BFGS_m'])
        self.__BFGS_maxfun = int(self.parameters['BFGS_maxfun'])
        self.__BFGS_factr = float(self.parameters['BFGS_factr'])
        # This is a hack for 64 bit systems (Linux). The machine precision 
        # is different for the BFGS optimizer (Fortran code) and we fix this by:
        is_64bits = sys.maxsize > 2**32
        if is_64bits:
            logging.debug("64-bit system detected, modifying BFGS_factr!")
            self.__BFGS_factr = 0.000488288*self.__BFGS_factr
        self.__BFGS_pgtol = float(self.parameters['BFGS_pgtol'])
        self.__BFGS_verbose = int(self.parameters['BFGS_verbose'])
        self.__surrogate_gamma = float(self.parameters['surrogate_gamma'])
        self.__s = float(self.parameters['surrogate_s'])
        self.__breakpoint_for_exp = float(self.parameters['breakpoint_for_exp'])
        self.__b = self.__estimate_r

    def __optimize(self):
        logging.debug("Starting optimization with BFGS ...")
        self.__needed_function_calls = 0
        # starting_point
        c_current = np.zeros(self.__dim, np.float64)
        c_current[self.__dim-1] = self.__b
        # Annealing sequence.
        for i in range(len(self.__lam_Uvec)):
            self.__lamU = self.__lam_Uvec[i]
            # crop one dimension (in case the offset b is fixed)
            c_current = c_current[:self.__dim-1]
            c_current = self.__localSearch(c_current)
            # reappend it if needed
            c_current = np.append(c_current, self.__b)
        f_opt = self.__getFitness(c_current)
        return c_current, f_opt

    def __localSearch(self, start):
        c_opt, f_opt, d = optimize.fmin_l_bfgs_b(self.__getFitness, start, m=self.__BFGS_m, \
                                     fprime=self.__getFitness_Prime, maxfun=self.__BFGS_maxfun,\
                                     factr=self.__BFGS_factr, pgtol=self.__BFGS_pgtol, iprint=self.__BFGS_verbose)
        self.__needed_function_calls += int(d['funcalls'])
        return c_opt

    def __getFitness(self,c):
        # check whether the function is called from the bfgs solver 
        # (that does not optimize the offset b) or not
        if len(c) == self.__dim - 1:
            c = np.append(c, self.__b)
        c = np.mat(c)
        b = c[:,self.__dim-1].T
        c_new = c[:,0:self.__dim-1].T
        c_new_sum = np.sum(c_new)
        XTc = self.X_T*c_new - self.__mean_u.T*c_new_sum
        preds_labeled = self.__surrogate_gamma*(1.0 - np.multiply(self.__YL, (self.X_l*XTc - self.__mean_u*XTc) + b[0,0]))
        preds_unlabeled = (self.X_u*XTc - self.__mean_u*XTc)  + b[0,0]
        # This vector has a "one" for each "numerically instable" entry; "zeros" for "good ones". 
        preds_labeled_conflict_indicator = np.sign(np.sign(preds_labeled/self.__breakpoint_for_exp - 1.0) + 1.0)
        # This vector has a one for each good entry and zero otherwise
        preds_labeled_good_indicator = (-1)*(preds_labeled_conflict_indicator - 1.0)
        preds_labeled_for_conflicts = np.multiply(preds_labeled_conflict_indicator,preds_labeled) 
        preds_labeled = np.multiply(preds_labeled,preds_labeled_good_indicator)
        # Compute values for good entries
        preds_labeled_log_exp = np.log(1.0 + np.exp(preds_labeled))
        # Compute values for instable entries
        preds_labeled_log_exp = np.multiply(preds_labeled_good_indicator, preds_labeled_log_exp)
        # Replace critical values with values 
        preds_labeled_final = preds_labeled_log_exp + preds_labeled_for_conflicts
        term1 = (1.0/(self.__surrogate_gamma*self.__size_l)) * np.sum(preds_labeled_final)
        preds_unlabeled_squared = np.multiply(preds_unlabeled,preds_unlabeled)
        term2 = (float(self.__lamU)/float(self.__size_u))*np.sum(np.exp(-self.__s * preds_unlabeled_squared))
        term3 = self.__lam * c_new.T * (self.X * XTc - self.__mean_u*XTc)
        return (term1 + term2 + term3)[0,0]

    def __getFitness_Prime(self,c):
        # check whether the function is called from the bfgs solver 
        # (that does not optimize the offset b) or not
        if len(c) == self.__dim - 1:
            c = np.append(c, self.__b)
        c = np.mat(c)
        b = c[:,self.__dim-1].T
        c_new = c[:,0:self.__dim-1].T
        c_new_sum = np.sum(c_new)
        XTc = self.X_T*c_new - self.__mean_u.T*c_new_sum
        preds_labeled = self.__surrogate_gamma*(1.0 - np.multiply(self.__YL, (self.X_l*XTc -self.__mean_u*XTc) + b[0,0]))
        preds_unlabeled = (self.X_u*XTc - self.__mean_u*XTc )+ b[0,0]
        preds_labeled_conflict_indicator = np.sign(np.sign(preds_labeled/self.__breakpoint_for_exp - 1.0) + 1.0)
        # This vector has a one for each good entry and zero otherwise
        preds_labeled_good_indicator = (-1)*(preds_labeled_conflict_indicator - 1.0)
        preds_labeled = np.multiply(preds_labeled,preds_labeled_good_indicator)
        preds_labeled_exp = np.exp(preds_labeled)
        term1 = np.multiply(preds_labeled_exp, 1.0/(1.0 + preds_labeled_exp))
        term1 = np.multiply(preds_labeled_good_indicator, term1)
        # Replace critical values with "1.0"
        term1 = term1 + preds_labeled_conflict_indicator
        term1 = np.multiply(self.__YL, term1)
        preds_unlabeled_squared_exp_f = np.multiply(preds_unlabeled,preds_unlabeled)
        preds_unlabeled_squared_exp_f = np.exp(-self.__s * preds_unlabeled_squared_exp_f)
        preds_unlabeled_squared_exp_f = np.multiply(preds_unlabeled_squared_exp_f, preds_unlabeled)
        term1_sum = np.sum(term1)
        tmp = self.X_l_T * term1 - self.__mean_u.T*term1_sum
        term1 = (-1.0/self.__size_l) * (self.X * tmp - self.__mean_u*tmp)
        preds_unlabeled_squared_exp_f_sum = np.sum(preds_unlabeled_squared_exp_f)
        tmp_unlabeled = self.X_u_T * preds_unlabeled_squared_exp_f - self.__mean_u.T * preds_unlabeled_squared_exp_f_sum
        term2 = ((-2.0 * self.__s * self.__lamU)/float(self.__size_u)) * (self.X * tmp_unlabeled - self.__mean_u*tmp_unlabeled)
        XTc_sum = np.sum(XTc)
        term3 = 2*self.__lam*(self.X * XTc - self.__mean_u*XTc)
        return np.array((term1 + term2 + term3).T)[0]

    def __recomputeModel(self, indi):
        self.__c = np.mat(indi[0]).T

############################################################################################
############################################################################################
class LinearKernel():
    """
    Linear Kernel
    """
    def __init__(self):
        pass

    def computeKernelMatrix(self, data1, data2, symmetric=False):
        """
        Computes the kernel matrix
        """
        logging.debug("Starting Linear Kernel Matrix Computation...")
        self._data1 = np.mat(data1)
        self._data2 = np.mat(data2)
        # print(self._data1.shape[1])
        # print((self._data2.T).shape[0])


        assert self._data1.shape[1] == (self._data2.T).shape[0]
        try:
            return self._data1 * self._data2.T
        except Exception as e:
            logging.error("Error while computing kernel matrix: " + str(e))
            import traceback
            traceback.print_exc()
            sys.exit()
        logging.debug("Kernel Matrix computed...")

    def getKernelValue(self, xi, xj):
        """
        Returns a single kernel value.
        """
        xi = np.array(xi)
        xj = np.array(xj)
        val = np.dot(xi, xj)
        return val


class DictLinearKernel():
    """
    Linear Kernel (for dictionaries)
    """
    def __init__(self):
        pass

    def computeKernelMatrix(self, data1, data2, symmetric=False):
        """
        Computes the kernel matrix
        """
        logging.debug("Starting Linear Kernel Matrix Computation...")
        self._data1 = data1
        self._data2 = data2
        self._dim1 = len(data1)
        self._dim2 = len(data2)
        self._symmetric = symmetric
        self.__km = None
        try:
            km = np.mat(np.zeros((self._dim1, self._dim2), dtype=np.float64))
            if self._symmetric:
                for i in range(self._dim1):
                    message = 'Kernel Matrix Progress: %dx%d/%dx%d' % (i, self._dim2,self._dim1,self._dim2)
                    logging.debug(message)
                    for j in range(i, self._dim2):
                        val = self.getKernelValue(self._data1[i], self._data2[j])
                        km[i, j] = val
                        km[j, i] = val
                return km
            else:
                for i in range(self._dim1):
                    message = 'Kernel Matrix Progress: %dx%d/%dx%d' % (i, self._dim2,self._dim1,self._dim2)
                    logging.debug(message)
                    for j in range(0, self._dim2):
                        val = self.getKernelValue(self._data1[i], self._data2[j])
                        km[i, j] = val
                return km
            
        except Exception as e:
            logging.error("Error while computing kernel matrix: " + str(e))
            sys.exit()
        logging.debug("Kernel Matrix computed...")

    def getKernelValue(self, xi, xj):
        """
        Returns a single kernel value.
        """
        val = 0.
        for key in xi:
            if key in xj:
                val += xi[key]*xj[key]
        return val

class RBFKernel():
    """
    RBF Kernel
    """
    def __init__(self, sigma):
        self.__sigma = sigma
        self.__sigma_squared_inv = 1.0 / (2* (self.__sigma ** 2) )

    def computeKernelMatrix(self, data1, data2, symmetric=False):
        """
        Computes the kernel matrix
        """
        logging.debug("Starting RBF Kernel Matrix Computation...")
        self._data1 = np.mat(data1)
        self._data2 = np.mat(data2)
        # import ipdb
        # ipdb.set_trace()
        # print(self._data1.shape[1])
        # print((self._data2.T).shape[0])

        assert self._data1.shape[1] == (self._data2.T).shape[0]
        self._dim1 = len(data1)
        self._dim2 = len(data2)
        self._symmetric = symmetric
        self.__km = None
        try:
            if self._symmetric:
                linearkm = self._data1 * self._data2.T
                trnorms = np.mat(np.diag(linearkm)).T
                trace_matrix = trnorms * np.mat(np.ones((1, self._dim1), dtype = np.float64))
                self.__km = trace_matrix + trace_matrix.T
                self.__km = self.__km - 2*linearkm
                self.__km = - self.__sigma_squared_inv * self.__km
                self.__km = np.exp(self.__km)
                return self.__km   
            else:
                m = self._data1.shape[0]
                n = self._data2.shape[0]
                assert self._data1.shape[1] == self._data2.shape[1]
                linkm = np.mat(self._data1 * self._data2.T)
                trnorms1 = []
                for i in range(m):
                    trnorms1.append((self._data1[i] * self._data1[i].T)[0,0])
                trnorms1 = np.mat(trnorms1).T
                trnorms2 = []
                for i in range(n):
                    trnorms2.append((self._data2[i] * self._data2[i].T)[0,0])
                trnorms2 = np.mat(trnorms2).T
                self.__km = trnorms1 * np.mat(np.ones((n, 1), dtype = np.float64)).T
                self.__km = self.__km + np.mat(np.ones((m, 1), dtype = np.float64)) * trnorms2.T
                self.__km = self.__km - 2 * linkm
                self.__km = - self.__sigma_squared_inv * self.__km
                self.__km = np.exp(self.__km)
                return self.__km
        except Exception as e:
            logging.error("Error while computing kernel matrix: " + str(e))
            sys.exit()

    def getKernelValue(self, xi, xj):
        """
        Returns a single kernel value.
        """
        xi = np.array(xi)
        xj = np.array(xj)
        diff = xi-xj
        val = np.exp(-self.__sigma_squared_inv * (np.dot(diff, diff)))
        return val

class DictRBFKernel():
    """
    RBF Kernel (for dictionaries)
    """
    def __init__(self, sigma):
        self.__sigma = sigma
        self.__sigma_squared_inv = 1.0 / ((self.__sigma ** 2))

    def computeKernelMatrix(self, data1, data2, symmetric=False):
        """
        Computes the kernel matrix
        """
        logging.debug("Starting RBF Kernel Matrix Computation...")
        self._data1 = data1
        self._data2 = data2
        self._dim1 = len(data1)
        self._dim2 = len(data2)
        self._symmetric = symmetric
        self.__km = None
        try:
            km = np.mat(np.zeros((self._dim1, self._dim2), dtype=np.float64))
            if self._symmetric:
                for i in range(self._dim1):
                    message = 'Kernel Matrix Progress: %dx%d/%dx%d' % (i, self._dim2,self._dim1,self._dim2)
                    logging.debug(message)
                    for j in range(i, self._dim2):
                        val = self.getKernelValue(self._data1[i], self._data2[j])
                        km[i, j] = val
                        km[j, i] = val
                return km
            else:
                for i in range(0, self._dim1):
                    message = 'Kernel Matrix Progress: %dx%d/%dx%d' % (i, self._dim2,self._dim1,self._dim2)
                    logging.debug(message)
                    for j in range(0, self._dim2):
                        val = self.getKernelValue(self._data1[i], self._data2[j])
                        km[i, j] = val
                return km
        except Exception as e:
            logging.error("Error while computing kernel matrix: " + str(e))
            sys.exit()
        logging.info("Kernel Matrix computed...")

    def getKernelValue(self, xi, xj):
        """
        Returns a single kernel value.
        """
        diff = xi.copy()
        for key in xj:
            if key in diff:
                diff[key]-=xj[key]
            else:
                diff[key]=-xj[key]
        diff = diff.values()
        val = np.exp(-self.__sigma_squared_inv * (np.dot(diff, diff)))
        return val


from __future__ import annotations
import collections
from typing_extensions import Self

from instancelib.instances.base import Instance

import functools
import itertools
import logging
from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    Deque,
    Dict,
    FrozenSet,
    Generic,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    Set,
)


from ..environment.base import AbstractEnvironment, IT

DT = TypeVar("DT")
VT = TypeVar("VT")
KT = TypeVar("KT")
LT = TypeVar("LT")
RT = TypeVar("RT")
LVT = TypeVar("LVT")
PVT = TypeVar("PVT")
FT = TypeVar("FT")
F = TypeVar("F", bound=Callable[..., Any])

ProbabilityPrediction = FrozenSet[Tuple[LT, float]]
LabelPrediction = FrozenSet[LT]

LOGGER = logging.getLogger(__name__)


class ActiveLearner(ABC, Iterator[IT], Generic[IT, KT, DT, VT, RT, LT]):
    """The **Abstract Base Class** `ActiveLearner` specifies the design for all
    Active Learning algorithms.

    Examples
    --------
    Instances can be sampled as follows

    >>> instance = next(al)

    Or in batch mode by using :func:`itertools.islice`:

    >>> instances = itertools.islice(al, 10) # Change the number to get more instances

    Mark a document as labeled

    >>> al.set_as_labeled(instance)

    Mark a document as unlabeled

    >>> al.set_as_unlabeled(instance)

    Update the ordering

    >>> al.update_ordering()

    Check how many documents are labeled

    >>> al.len_labeled
    """

    _name = "ActiveLearner"
    """The internal name for this Active Learner"""
    ordering: Optional[Deque[KT]]
    """The ordering of the Active Learner"""
    env: AbstractEnvironment[IT, KT, DT, VT, RT, LT]
    """The internal environment"""
    identifier: Optional[str]
    """The identifier of the Active Learning method"""

    @property
    def name(self) -> Tuple[str, Optional[LT]]:
        """Return the name of the Active Learner

        Returns
        -------
        Tuple[str, Optional[LT]]
            The tuple contains a name and optionally the label if it the learner
            optimizes for a specific label
        """
        if self.identifier is not None:
            return self.identifier, None
        return self._name, None

    def __iter__(self) -> Self:
        """The Active Learning class is an iterator, iterating
        over instances

        Returns
        -------
        ActiveLearner[IT, KT, DT, VT, RT, LT]
            The same Active Learner is already an iterator, so ``iter(al) == al``
        """
        return self

    @abstractmethod
    def update_ordering(self) -> bool:
        """Update the ordering of the Active Learning method

        Returns
        -------
        bool
            True if updating the ordering succeeded
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def has_ordering(self) -> bool:
        """Returns true if an ordering has been established for this Active Learner

        Returns
        -------
            bool: True if an ordering has been established

        See Also
        --------
        update_ordering : The method to create or update the ordering
        """
        raise NotImplementedError

    @abstractmethod
    def __next__(self) -> IT:
        """Return the next instance based on the ordering

        Returns
        -------
        Instance[KT, DT, VT, RT]
            The most informative instance based on the learners ordering

        See Also
        --------
        __iter__ : Optional function for iterating over instances

        Examples
        --------
        `ActiveLearner` objects can be used as follows for retrieving instances:

        >>> # Initialize an ActiveLearner object
        >>> al = ActiveLearner()
        >>> # Attach an environment
        >>> al = al(env)
        >>> # Request the most informative instance
        >>> ins = next(al)
        >>> # Request the 10 most informative instances
        >>> inss = itertools.islice(al, 10)
        """
        raise NotImplementedError

    @staticmethod
    def iterator_log(func: F) -> F:
        """A decorator that logs iterator calls

        Parameters
        ----------
        func : F
            The ``__next__()`` function that iterates over Instances

        Returns
        -------
        F
            The same function with a logger wrapped around it
        """

        @functools.wraps(func)
        def wrapper(
            self: ActiveLearner[IT, KT, DT, VT, RT, LT],
            *args: Any,
            **kwargs: Dict[str, Any],
        ) -> F:
            result: Union[Any, IT] = func(self, *args, **kwargs)
            if isinstance(result, Instance):
                LOGGER.info(
                    "Sampled document %i with method %s", result.identifier, self.name
                )  # type: ignore
                self.env.logger.log_sample(result, self.name)
            return result  # type: ignore

        return wrapper  # type: ignore

    @staticmethod
    def label_log(func: F) -> F:
        """A decorator that logs label calls

        Parameters
        ----------
        func : F
            The function that labels an instance

        Returns
        -------
        F
            The same function with a logger wrapped around it
        """

        @functools.wraps(func)
        def wrapper(
            self: ActiveLearner[IT, KT, DT, VT, RT, LT],
            instance: Instance[KT, DT, VT, RT],
            *args: Any,
            **kwargs: Any,
        ):
            labels = self.env.labels.get_labels(instance)
            self.env.logger.log_label(instance, self.name, *labels)
            return func(self, instance, *args, **kwargs)

        return wrapper  # type: ignore

    @staticmethod
    def ordering_log(func: F) -> F:
        """A decorator that logs `ordering function` calls

        Parameters
        ----------
        func : F
            The function that establishes an ordering

        Returns
        -------
        F
            The same function with a logger wrapped around it
        """

        @functools.wraps(func)
        def wrapper(self: ActiveLearner, *args: Any, **kwargs: Dict[str, Any]) -> F:
            ordering, ordering_metric = func(self, *args, **kwargs)
            self.env.logger.log_ordering(
                ordering, ordering_metric, self.env.labeled.key_list, self.env.labels
            )
            return ordering, ordering_metric  # type: ignore

        return wrapper  # type: ignore

    def query(self) -> Optional[IT]:
        """Query the most informative instance

        Returns
        -------
        Optional[Instance[KT, DT, VT, RT]]
            The most informative `Instance`.
            It will return ``None`` if there are no more documents
        """
        return next(self, None)

    def query_batch(self, batch_size: int) -> Sequence[IT]:
        """Query the `batch_size` most informative instances

        Parameters
        ----------
        batch_size : int
            The size of the batch

        Returns
        -------
        Sequence[Instance[KT, DT, VT, RT]]
            A batch with ``len(batch) <= batch_size``
        """
        return list(itertools.islice(self, batch_size))

    @abstractmethod
    def set_as_labeled(self, instance: Instance[KT, DT, VT, RT]) -> None:
        """Mark the instance as labeled

        Parameters
        ----------
        instance : Instance[KT, DT, VT, RT]
            The now labeled instance
        """
        raise NotImplementedError

    @abstractmethod
    def set_as_unlabeled(self, instance: Instance[KT, DT, VT, RT]) -> None:
        """Mark the instance as unlabeled

        Parameters
        ----------
        instance : Instance[KT, DT, VT, RT]
            The now labeled instance
        """
        raise NotImplementedError

    @property
    def len_unlabeled(self) -> int:
        """Return the number of unlabeled documents

        Returns
        -------
        int
            The number of labeled documents
        """
        return len(self.env.unlabeled)

    @property
    def len_labeled(self) -> int:
        """Return the number of labeled documents

        Returns
        -------
        int
            The number of labeled documents
        """
        return len(self.env.labeled)

    @property
    def size(self) -> int:
        """The number of initial unlabeled documents

        Returns
        -------
        int
            The number of unlabeled documents
        """
        return self.len_labeled + self.len_unlabeled

    @property
    def ratio_learned(self) -> float:
        """The labeling progress sofar

        Returns
        -------
        float
            The ratio of labeled documents; compared to the
        """
        return self.len_labeled / self.size

    @classmethod
    @abstractmethod
    def builder(cls, *args: Any, **kwargs: Any) -> Callable[..., Self]:
        raise NotImplementedError



from __future__ import annotations

import collections
from math import ceil
from typing import (
    Any,
    Callable,
    Deque,
    Dict,
    Generic,
    Mapping,
    Optional,
    Sequence,
    Tuple,
)
from typing_extensions import Self
from uuid import uuid4

import instancelib as il
import numpy as np
import numpy.typing as npt
from instancelib.ingest.qrel import TrecDataset
from instancelib.typehints import DT, KT, LT, RT, VT

from ..analysis.base import AbstractStatistics, AnnotationStatistics, StatsMixin
from ..analysis.initialization import Initializer
from ..environment.base import AbstractEnvironment
from ..typehints import IT
from ..utils.func import list_unzip, sort_on
from ..utils.numpy import raw_proba_chainer
from .base import ActiveLearner
from .poolbased import PoolBasedAL

PSEUDO_INS_PROVIDER = "PSEUDO_INSTANCES"


class PseudoInstanceInitializer(Initializer[IT, KT, LT], Generic[IT, KT, DT, LT]):
    def __init__(self, pseudo_data: DT, pseudo_label: LT):
        self.pseudo_data = pseudo_data
        self.pseudo_label = pseudo_label
        self.pseudo_id = uuid4()

    def __call__(
        self, learner: ActiveLearner[IT, KT, DT, VT, RT, LT]
    ) -> ActiveLearner[IT, KT, DT, VT, RT, LT]:
        ins = learner.env.create(data=self.pseudo_data, vector=None)
        learner.env.create_named_provider(PSEUDO_INS_PROVIDER, [ins.identifier])
        return learner

    @classmethod
    def from_trec(
        cls, trec: TrecDataset, topic_id: str
    ) -> PseudoInstanceInitializer[IT, str, str, str]:
        topic_df = trec.topics.set_index("id")
        topic_row = topic_df.xs(topic_id)
        pseudo_data: str = topic_row.title + " " + topic_row.query  # type: ignore
        return cls(pseudo_data, trec.pos_label)  # type: ignore


class AutoTarLearner(
    PoolBasedAL[IT, KT, DT, VT, RT, LT],
    StatsMixin[KT, LT],
    Generic[IT, KT, DT, VT, RT, LT],
):
    rank_history: Dict[int, Mapping[KT, int]]
    sampled_sets: Dict[int, Sequence[KT]]
    batch_sizes: Dict[int, int]
    current_sample: Deque[KT]
    rng: np.random.Generator
    _stats: AbstractStatistics[KT, LT]

    def __init__(
        self,
        env: AbstractEnvironment[IT, KT, DT, VT, RT, LT],
        classifier: Callable[
            [AbstractEnvironment[IT, KT, DT, VT, RT, LT]],
            il.AbstractClassifier[
                IT, KT, DT, VT, RT, LT, npt.NDArray[Any], npt.NDArray[Any]
            ],
        ],
        pos_label: LT,
        neg_label: LT,
        k_sample: int,
        batch_size: int,
        *_,
        seed: int = 0,
        identifier: Optional[str] = None,
        **__,
    ) -> None:
        super().__init__(env, *_, identifier=identifier, **__)
        # Problem definition
        self.classifier = classifier(env)
        self.pos_label = pos_label
        self.neg_label = neg_label

        # Batch and sample sizes
        self.k_sample = k_sample
        self.batch_sizes = dict()
        self.batch_size = batch_size

        # Record keeping for the current sample
        self.it = 0
        self.current_sample = collections.deque()
        self.batch_sizes[self.it] = batch_size

        # Record keeping for recall analysis
        self.rank_history = dict()
        self.sampled_sets = dict()

        # Random generator for sampling
        self.rng = np.random.default_rng(seed)

        # Statistics Logger for Stopping
        self._stats = AnnotationStatistics()

    @property
    def stats(self) -> AbstractStatistics[KT, LT]:
        return self._stats

    def update_ordering(self) -> bool:
        return True

    def _provider_sample(
        self, provider: il.InstanceProvider[IT, KT, DT, VT, RT]
    ) -> il.InstanceProvider[IT, KT, DT, VT, RT]:
        k_sample = min(self.k_sample, len(provider))
        sampled_keys: Sequence[KT] = self.rng.choice(
            provider.key_list, size=k_sample, replace=False  # type: ignore
        ).tolist()
        sampled_provider = self.env.create_bucket(sampled_keys)
        return sampled_provider

    def _temp_augment_and_train(self):
        temp_labels = il.MemoryLabelProvider[KT, LT].from_provider(self.env.labels)
        sampled_non_relevant = self._provider_sample(self.env.unlabeled)
        if PSEUDO_INS_PROVIDER in self.env:
            pseudo_docs = self.env[PSEUDO_INS_PROVIDER]
            for ins_key in pseudo_docs:
                temp_labels.set_labels(ins_key, self.pos_label)
        else:
            pseudo_docs = self.env.create_bucket([])
        for ins_key in sampled_non_relevant:
            temp_labels.set_labels(ins_key, self.neg_label)
        train_set = self.env.combine(
            sampled_non_relevant, self.env.labeled, pseudo_docs
        )
        self.classifier.fit_provider(train_set, temp_labels)

    def _predict(
        self, provider: il.InstanceProvider[IT, KT, DT, VT, RT]
    ) -> Tuple[Sequence[KT], npt.NDArray[Any]]:
        raw_probas = self.classifier.predict_proba_provider_raw(provider)
        keys, matrix = raw_proba_chainer(raw_probas)
        return keys, matrix

    def _rank(
        self, provider: il.InstanceProvider[IT, KT, DT, VT, RT]
    ) -> Sequence[Tuple[KT, float]]:
        keys, matrix = self._predict(provider)
        pos_column = self.classifier.get_label_column_index(self.pos_label)
        prob_vector = matrix[:, pos_column]
        floats: Sequence[float] = prob_vector.tolist()
        zipped = list(zip(keys, floats))
        ranking = sort_on(1, zipped)
        return ranking

    def _sample(self, distribution: Sequence[Tuple[KT, float]]) -> Sequence[KT]:
        keys, _ = list_unzip(distribution)
        sample = keys[0 : self.batch_size]
        return sample

    @classmethod
    def _to_history(cls, ranking: Sequence[Tuple[KT, float]]) -> Mapping[KT, int]:
        keys, _ = list_unzip(ranking)
        history = {k: i for i, k in enumerate(keys, start=1)}
        return history

    def update_sample(self) -> Deque[KT]:
        if not self.current_sample:
            self.stats.update(self)
            self._temp_augment_and_train()
            ranking = self._rank(self.env.unlabeled)
            sample = self._sample(ranking)

            # Store current sample
            self.current_sample = collections.deque(sample)

            # Record keeping
            self.rank_history[self.it] = self._to_history(ranking)
            self.sampled_sets[self.it] = tuple(sample)
            self.batch_sizes[self.it] = self.batch_size

            # Increment batch_size for next for train iteration
            self.batch_size += ceil(self.batch_size / 10)
            self.it += 1
        return self.current_sample

    def __next__(self) -> IT:
        if self.env.unlabeled.empty:
            raise StopIteration
        self.update_sample()
        while self.current_sample:
            ins_key = self.current_sample.popleft()
            if ins_key not in self.env.labeled:
                return self.env.dataset[ins_key]
        if not self.env.unlabeled.empty:
            return self.__next__()
        raise StopIteration()

    @classmethod
    def builder(
        cls,
        classifier: Callable[
            [AbstractEnvironment[IT, KT, DT, VT, RT, LT]],
            il.AbstractClassifier[
                IT, KT, DT, VT, RT, LT, npt.NDArray[Any], npt.NDArray[Any]
            ],
        ],
        k_sample: int,
        batch_size: int,
    ) -> Callable[..., Self]:
        def builder_func(
            env: AbstractEnvironment[IT, KT, DT, VT, RT, LT],
            pos_label: LT,
            neg_label: LT,
            *_,
            **__,
        ):
            classifier(env)
            return cls(env, classifier, pos_label, neg_label, k_sample, batch_size)

        return builder_func



# # Copyright 2019 The ASReview Authors. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License");
# # you may not use this file except in compliance with the License.
# # You may obtain a copy of the License at
# #
# #     http://www.apache.org/licenses/LICENSE-2.0
# #
# # Unless required by applicable law or agreed to in writing, software
# # distributed under the License is distributed on an "AS IS" BASIS,
# # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# # See the License for the specific language governing permissions and
# # limitations under the License.
# from typing import Optional

# import logging

# import numpy as np # type: ignore

# from .base import BaseBalancer
# from .double import _one_weight
# from .double import DoubleBalancer
# from .double import _zero_weight
# from .double import fill_training
# from .double import random_round
# from ..utils import get_random_generator 

# from ..activelearning.catalog import ALCatalog

# LOGGER = logging.getLogger(__name__)

# class TripleBalancer(BaseBalancer):
#     """
#     Class to get the three way rebalancing function and arguments.
#     It divides the data into three groups: 1's, 0's from random sampling,
#     and 0's from max sampling. Thus it only makes sense to use this class in
#     combination with the rand_max query strategy.
#     """
#     name = "triple"

#     def __init__(self,
#                  a: float=2.155,
#                  alpha: float=0.94,
#                  b: float=0.789,
#                  beta: float=1.0,
#                  c: float=0.835,
#                  gamma: float=2.0,
#                  shuffle: bool=True,
#                  random_state: Optional[np.random.Generator] = None): # type: ignore
#         """Initialize the triple balance strategy.

#         Arguments
#         ---------
#         a: float
#             Governs the weight of the 1's. Higher values mean linearly more 1's
#             in your training sample.
#         alpha: float
#             Governs the scaling the weight of the 1's, as a function of the
#             ratio of ones to zeros. A positive value means that the lower the
#             ratio of zeros to ones, the higher the weight of the ones.
#         b: float
#             Governs how strongly we want to sample depending on the total
#             number of samples. A value of 1 means no dependence on the total
#             number of samples, while lower values mean increasingly stronger
#             dependence on the number of samples.
#         beta: float
#             Governs the scaling of the weight of the zeros depending on the
#             number of samples. Higher values means that larger samples are more
#             strongly penalizing zeros.
#         c: float
#             Value between one and zero that governs the weight of samples done
#             with maximal sampling. Higher values mean higher weight.
#         gamma: float
#             Governs the scaling of the weight of the max samples as a function
#             of the % of papers read. Higher values mean stronger scaling.
#         """
#         super(TripleBalancer, self).__init__()
#         self.a = a
#         self.alpha = alpha
#         self.b = b
#         self.beta = beta
#         self.c = c
#         self.gamma = gamma
#         self.shuffle = shuffle
#         self.fallback_model = DoubleBalancer(a=a, alpha=alpha, b=b, beta=beta,
#                                             rng=random_state)
#         self._random_state = get_random_generator(random_state)
    
#     def resample(self, X, y):
#         max_idx = self._environment.logger.get_instances_by_method(ALCatalog.QueryType.LABELMAXIMIZER)
#         max_idx = np.array(shared["query_src"].get("max", []), dtype=np.int)
#         rand_idx = np.array([], dtype=np.int)
#         for qtype in shared["query_src"]:
#             if qtype != "max":
#                 rand_idx = np.append(rand_idx, shared["query_src"][qtype])

#         rand_idx = rand_idx.astype(int)
#         # Write them back for next round.
#         if self.shuffle:
#             self._random_state.shuffle(rand_idx)
#             self._random_state.shuffle(max_idx)

#         if len(rand_idx) == 0 or len(max_idx) == 0:
#             LOGGER.debug("Warning: trying to use triple balance, but unable"
#                           f" to, because we have {len(max_idx)} max samples "
#                           f"and {len(rand_idx)} random samples.")
#             return self.fallback_model.resample(X, y)

#         # Split the idx into three groups: 1's, random 0's, max 0's.
#         one_idx = train_idx[np.where(y[train_idx] == 1)]
#         zero_max_idx = max_idx[np.where(y[max_idx] == 0)]
#         zero_rand_idx = rand_idx[np.where(y[rand_idx] == 0)]

#         if len(zero_rand_idx) == 0 or len(zero_max_idx) == 0:
#             LOGGER.debug("Warning: trying to use triple balance, but unable "
#                           f"to, because we have {len(zero_max_idx)} zero max"
#                           f"samples and {len(zero_rand_idx)} random samples.")
#             return self.fallback_model.sample(X, y, train_idx, shared)

#         n_one = len(one_idx)
#         n_zero_rand = len(zero_rand_idx)
#         n_zero_max = len(zero_max_idx)
#         n_samples = len(y)
#         n_train = len(train_idx)

#         # Get the distribution of 1's, and random 0's and max 0's.
#         n_one_train, n_zero_rand_train, n_zero_max_train = _get_triple_dist(
#             n_one, n_zero_rand, n_zero_max, n_samples, n_train, self.a,
#             self.alpha, self.b, self.beta, self.c, self.gamma,
#             self._random_state)
#         logging.debug(f"(1, 0_rand, 0_max) = ({n_one_train}, "
#                       f"{n_zero_rand_train}, {n_zero_max_train})")

#         one_train_idx = fill_training(one_idx, n_one_train, self._random_state)
#         zero_rand_train_idx = fill_training(zero_rand_idx, n_zero_rand_train,
#                                             self._random_state)
#         zero_max_train_idx = fill_training(zero_max_idx, n_zero_max_train,
#                                            self._random_state)

#         all_idx = np.concatenate(
#             [one_train_idx, zero_rand_train_idx, zero_max_train_idx])
#         self._random_state.shuffle(all_idx)

#         return X[all_idx], y[all_idx]

#     def full_hyper_space(self):
#         from hyperopt import hp
#         parameter_space = {
#             "bal_a": hp.lognormal("bal_a", 0, 1),
#             "bal_alpha": hp.uniform("bal_alpha", 0, 2),
#             "bal_b": hp.uniform("bal_b", 0, 1),
#             # "bal_zero_beta": hp.uniform("bal_zero_beta", 0, 2),
#             "bal_c": hp.uniform("bal_c", 0, 1),
#             # "bal_zero_max_gamma": hp.uniform("bal_zero_max_gamma", 0.01, 2)
#         }
#         return parameter_space, {}


# def _zero_max_weight(fraction_read, c, gamma):
#     """
#     Get the weight ratio between ones and zeros.

#     Parameters
#     ----------
#     beta:
#         Exponent governing decay of 1's to 0's.
#     delta:
#         Asymptotic ratio for infinite number of samples.

#     Returns
#     -------
#     float:
#         Weight ratio between 1's and 0's
#     """
#     weight = 1 - (1 - c) * (1 - fraction_read)**gamma
#     return weight


# def _get_triple_dist(n_one, n_zero_rand, n_zero_max, n_samples, n_train, one_a,
#                      one_alpha, zero_b, zero_beta, zero_max_c, zero_max_gamma,
#                      random_state):
#     " Get the number of 1's, random 0's and max 0's in each mini epoch. "
#     n_zero = n_zero_rand + n_zero_max
#     n_read = n_one + n_zero
#     one_weight = _one_weight(n_one, n_zero, one_a, one_alpha)
#     zero_weight = _zero_weight(n_read, zero_b, zero_beta)
#     zero_max_weight = _zero_max_weight(n_read / n_samples, zero_max_c,
#                                        zero_max_gamma)

#     tot_zo_weight = one_weight * n_one + zero_weight * n_zero

#     n_one_train = random_round(one_weight * n_one * n_train / tot_zo_weight,
#                                random_state)
#     n_one_train = max(1, min(n_train - 2, n_one_train))
#     n_zero_train = n_train - n_one_train

#     tot_rm_weight = 1 * n_zero_rand + zero_max_weight * n_zero_max
#     n_zero_rand_train = random_round(
#         n_zero_train * 1 * n_zero_rand / tot_rm_weight, random_state)
#     n_zero_rand_train = max(1, min(n_zero_rand - 1, n_zero_rand_train))
#     n_zero_max_train = n_zero_train - n_zero_rand_train

#     return n_one_train, n_zero_rand_train, n_zero_max_train



# from __future__ import annotations

# from typing import Sequence, Any, Generic, TypeVar

# import numpy as np # type: ignore

# from instancelib.instances.base import ContextInstance
# from ..utils.func import list_unzip

# from .base import BaseVectorizer, SeparateContextVectorizer



# DT = TypeVar("DT")
# CT = TypeVar("CT")
# class ContextVectorizer(BaseVectorizer[ContextInstance[Any, DT, npt.NDArray[Any], Any, CT]], Generic[DT, CT]):
#     _name = "ContextVectorizer"
#     def __init__(self,
#                  vectorizer: SeparateContextVectorizer[DT, CT]) -> None:
#         super().__init__()
#         self.innermodel = vectorizer

#     def fitted(self) -> bool:
#         return self.innermodel.fitted

#     def fit(self, 
#             x_data: Sequence[ContextInstance[Any, DT, npt.NDArray[Any], Any, CT]], 
#             **kwargs: Any) -> ContextVectorizer[DT, CT]: # type: ignore
#         texts, contexts = list_unzip(((x.data, x.context) for x in x_data))
#         self.innermodel.fit(texts, contexts, **kwargs)
#         return self

#     def transform(self, 
#                   x_data: Sequence[ContextInstance[Any, DT, npt.NDArray[Any], Any, CT]], 
#                   **kwargs: Any) -> npt.NDArray[Any]: # type: ignore
#         texts, contexts = list_unzip(((x.data, x.context) for x in x_data))
#         return self.innermodel.transform(texts, contexts, **kwargs)

#     def fit_transform(self, 
#                       x_data: Sequence[ContextInstance[Any, DT, npt.NDArray[Any], Any, CT]], 
#                       **kwargs: Any) -> npt.NDArray[Any]: # type: ignore
#         self.fit(x_data, **kwargs) # type: ignore
#         return self.transform(x_data, **kwargs) # type: ignore



from __future__ import annotations

import collections
import math
from typing import Any, Callable, Deque, Generic, Optional

import numpy as np
import math
from instancelib.utils.func import all_equal, intersection

from ..activelearning import ActiveLearner
from ..activelearning.estimator import Estimator
from ..estimation.base import AbstractEstimator
from ..typehints import LT
from .base import AbstractStopCriterion
from .heuristic import AprioriRecallTarget, SameStateCount


class CaptureRecaptureCriterion(SameStateCount[LT], Generic[LT]):
    def __init__(
        self,
        calculator: AbstractEstimator[Any, Any, Any, Any, Any, LT],
        label: LT,
        same_state_count: int,
        margin: float,
    ):
        self.calculator = calculator
        super().__init__(label, same_state_count)
        self.estimate_history: Deque[float] = collections.deque()
        self.margin: float = margin

    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]):
        super().update(learner)
        if isinstance(learner, Estimator):
            self.add_count(learner.env.labels.document_count(self.label))
            estimate = self.calculator(learner, self.label)
            self.add_estimate(estimate.upper_bound)

    def add_estimate(self, value: float) -> None:
        if len(self.estimate_history) > self.same_state_count:
            self.estimate_history.pop()
        self.estimate_history.appendleft(value)

    @property
    def estimate(self) -> float:
        return self.estimate_history[0]

    @property
    def estimate_match(self) -> bool:
        difference = abs(self.estimate - self.count)
        return difference < self.margin

    @property
    def same_count(self) -> bool:
        return all_equal(self.pos_history)

    @property
    def same_estimate(self) -> bool:
        return all_equal(self.estimate_history)

    @property
    def stop_criterion(self) -> bool:
        if len(self.estimate_history):
            if len(self.pos_history) < self.same_state_count:
                return False
            return self.has_been_different and self.same_count and self.estimate_match
        return super().stop_criterion


class CaptureRecaptureCriterion2(CaptureRecaptureCriterion[LT], Generic[LT]):
    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]):
        self.add_count(learner.env.labels.document_count(self.label))
        if isinstance(learner, Estimator):
            estimate = self.calculator(learner, self.label)
            self.add_estimate(estimate.point)


class RaschCaptureCriterion(CaptureRecaptureCriterion[LT], Generic[LT]):
    @property
    def estimate(self) -> float:
        history = np.array([*self.estimate_history])
        return float(np.mean(history))

    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]):
        self.add_count(learner.env.labels.document_count(self.label))
        if isinstance(learner, Estimator):
            estimate = self.calculator(learner, self.label)
            dataset_size = len(learner.env.dataset)
            if estimate.point < dataset_size:
                self.add_estimate(estimate.point)


class EnsembleConvergenceCriterion(SameStateCount[LT], Generic[LT]):
    def __init__(self, label: LT, same_state_count: int, convergence_margin: float):
        super().__init__(label, same_state_count)
        self.margin = convergence_margin
        self.missing_history: Deque[float] = collections.deque()

    def add_missing(self, value: float) -> None:
        if len(self.missing_history) > self.same_state_count:
            self.missing_history.pop()
        self.missing_history.appendleft(value)

    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]):
        super().update(learner)
        if isinstance(learner, Estimator):
            common_positive = learner.env.labels.get_instances_by_label(self.label)
            positive_sets = intersection(
                *[
                    (
                        member.env.labels.get_instances_by_label(
                            self.label
                        ).intersection(member.env.labeled)
                    )
                    for member in learner.learners
                ]
            )
            missing_percentage = 1.0 - len(positive_sets) / len(common_positive)
            self.add_missing(missing_percentage)

    @property
    def missing_ratio(self) -> float:
        return self.missing_history[0]

    @property
    def within_margin(self) -> bool:
        return self.missing_ratio < self.margin

    @property
    def stop_criterion(self) -> bool:
        if self.missing_history:
            if len(self.pos_history) < self.same_state_count:
                return False
            return self.has_been_different and self.same_count and self.within_margin
        return super().stop_criterion


class CombinedStopCriterion(RaschCaptureCriterion[LT], Generic[LT]):
    def __init__(
        self,
        calculator: AbstractEstimator[Any, Any, Any, Any, Any, LT],
        label: LT,
        same_state_count: int,
        margin: float,
        convergence_margin: float,
    ):
        super().__init__(calculator, label, same_state_count, margin)
        self.convergence_margin = convergence_margin
        self.missing_history: Deque[float] = collections.deque()

    def add_missing(self, value: float) -> None:
        if len(self.missing_history) > self.same_state_count:
            self.missing_history.pop()
        self.missing_history.appendleft(value)

    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]):
        super().update(learner)
        if isinstance(learner, Estimator):
            common_positive = learner.env.labels.get_instances_by_label(self.label)

            positive_sets = intersection(
                *[
                    (
                        member.env.labels.get_instances_by_label(
                            self.label
                        ).intersection(member.env.labeled)
                    )
                    for member in learner.learners
                ]
            )
            missing_percentage = 1.0 - len(positive_sets) / len(common_positive)
            self.add_missing(missing_percentage)

    @property
    def estimate(self) -> float:
        return self.estimate_history[0]

    @property
    def missing_ratio(self) -> float:
        return self.missing_history[0]

    @property
    def within_margin(self) -> bool:
        return self.missing_ratio <= self.convergence_margin

    @property
    def stop_criterion(self) -> bool:
        if self.missing_history:
            if len(self.pos_history) < self.same_state_count:
                return False
            return (
                self.has_been_different
                and self.same_count
                and self.within_margin
                or super().stop_criterion
            )
        return super().stop_criterion


class TwoCombinedStopCriterion(CombinedStopCriterion, Generic[LT]):
    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]):
        super().update(learner)
        if isinstance(learner, Estimator):
            common_positive = learner.env.labels.get_instances_by_label(self.label)
            common_positives = np.array(
                ([len(common_positive) * len(learner.learners)])
            )
            positive_sets = np.array(
                [
                    len(
                        member.env.labels.get_instances_by_label(
                            self.label
                        ).intersection(member.env.labeled)
                    )
                    for member in learner.learners
                ]
            )
            missing_percentages = 1.0 - positive_sets / common_positives
            missing_percentage = 1.0 - len(positive_sets) / len(common_positive)
            self.add_missing(missing_percentage)


class UpperboundCombinedCritertion(CombinedStopCriterion[LT], Generic[LT]):
    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]):
        self.add_count(learner.env.labels.document_count(self.label))
        if isinstance(learner, Estimator):
            estimation = self.calculator(learner, self.label)
            dataset_size = len(learner.env.dataset)
            if estimation.upper_bound < dataset_size:
                self.add_estimate(estimation.upper_bound)
            common_positive = learner.env.labels.get_instances_by_label(self.label)
            positive_sets = intersection(
                *[
                    (
                        member.env.labels.get_instances_by_label(
                            self.label
                        ).intersection(member.env.labeled)
                    )
                    for member in learner.learners
                ]
            )
            missing_percentage = 1.0 - len(positive_sets) / len(common_positive)
            self.add_missing(missing_percentage)

    @property
    def estimate_match(self) -> bool:
        projected_recall = self.count / self.estimate
        return projected_recall >= 0.95


class RecallEstimation(AprioriRecallTarget):
    pass


class Conservative(AbstractStopCriterion[LT], Generic[LT]):
    estimate: Optional[float]

    def __init__(
        self,
        calculator: AbstractEstimator[Any, Any, Any, Any, Any, LT],
        label: LT,
        target: float,
    ):
        super().__init__()
        self.label = label
        self.calculator = calculator
        self.estimate = None
        self.count_found = 0
        self.target = target

    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]):
        self.count_found = learner.env.labels.document_count(self.label)
        estimate = self.get_estimate(learner)
        if estimate > len(learner.env.dataset) or estimate == float("nan"):
            self.estimate = None
        else:
            self.estimate = estimate

    def get_estimate(
        self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]
    ) -> float:
        return self.calculator(learner, self.label).upper_bound

    @classmethod
    def builder(
        cls, calculator: AbstractEstimator[Any, Any, Any, Any, Any, LT], target: float
    ) -> Callable[[LT, LT], Conservative[LT]]:
        def builder_func(pos_label: LT, neg_label: LT) -> Conservative[LT]:
            return cls(calculator, pos_label, target)

        return builder_func

    @property
    def stop_criterion(self) -> bool:
        if self.estimate is None:
            return False
        recall_estimate = self.count_found / self.estimate
        return round(recall_estimate, 2) >= self.target


class Optimistic(Conservative[LT], Generic[LT]):
    def get_estimate(
        self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]
    ) -> float:
        return self.calculator(learner, self.label).point



from __future__ import annotations

import logging
import random
from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)
from typing_extensions import Self

import instancelib as il
import numpy as np  # type: ignore
from instancelib import Instance, InstanceProvider
from instancelib.typehints.typevars import LMT, PMT

from ..activelearning.insclass import ILMLBased
from ..environment.base import IT, AbstractEnvironment
from ..machinelearning import AbstractClassifier
from ..utils import get_random_generator
from .base import ActiveLearner
from .ml_based import MLBased
from .poolbased import PoolBasedAL
from .random import RandomSampling

DT = TypeVar("DT")
VT = TypeVar("VT")
KT = TypeVar("KT")
LT = TypeVar("LT")
RT = TypeVar("RT")
LVT = TypeVar("LVT")
PVT = TypeVar("PVT")

LOGGER = logging.getLogger(__name__)


class AbstractEnsemble(ABC, Generic[IT, KT, DT, VT, RT, LT]):
    _name = "AbstractEnsemble"
    learners: Sequence[ActiveLearner[IT, KT, DT, VT, RT, LT]]
    env: AbstractEnvironment[IT, KT, DT, VT, RT, LT]
    _has_ordering: bool

    @abstractmethod
    def _choose_learner(self) -> ActiveLearner[IT, KT, DT, VT, RT, LT]:
        """Internal functions that selects the next active learner for the next query

        Returns:
            ActiveLearner[IT, KT, DT, VT, RT, LT]: One of the learners from the ensemble
        """
        raise NotImplementedError

    @property
    def has_ordering(self) -> bool:
        return self._has_ordering

    def update_ordering(self):
        """Updates the ordering for all learners of the ensemble"""
        for learner in self.learners:
            learner.update_ordering()
        self._has_ordering = True
        return True

    def __next__(self) -> IT:
        learner = self._choose_learner()
        return next(learner)

    @ActiveLearner.label_log
    def set_as_labeled(self, instance: Instance[KT, DT, VT, RT]) -> None:
        self.env.unlabeled.discard(instance)
        self.env.labeled.add(instance)

    def set_as_unlabeled(self, instance: Instance[KT, DT, VT, RT]) -> None:
        """Mark the instance as unlabeled

        Parameters
        ----------
        instance : Instance
            The now labeled instance
        """
        self.env.labeled.discard(instance)
        self.env.unlabeled.add(instance)


class ManualEnsemble(
    AbstractEnsemble[IT, KT, DT, VT, RT, LT],
    PoolBasedAL[IT, KT, DT, VT, RT, LT],
    Generic[IT, KT, DT, VT, RT, LT],
):
    _name = "Ensemble"

    def __init__(
        self,
        env: AbstractEnvironment[IT, KT, DT, VT, RT, LT],
        learners: Sequence[
            ActiveLearner[IT, KT, DT, VT, RT, LT],
        ],
        probabilities: Sequence[float],
        rng: Any = None,
        *_,
        identifier: Optional[str] = None,
        **__,
    ) -> None:
        super().__init__(env, identifier=identifier)
        self.learners = learners
        self.probabilities = probabilities
        self._sample_dict: Dict[KT, int] = {}
        self._rng: Any = get_random_generator(rng)

    def _choose_learner(self) -> ActiveLearner[IT, KT, DT, VT, RT, LT]:
        """Internal functions that selects the next active learner for the next query

        Returns
        -------
        ActiveLearner[IT, KT, DT, VT, RT, LT]
            One of the learners from the ensemble
        """
        idxs = np.arange(len(self.learners))
        al_idx: int = self._rng.choice(idxs, size=1, p=self.probabilities)[0]
        learner = self.learners[al_idx]
        return learner

    def __next__(self) -> IT:
        # Select the learner
        learner = self._choose_learner()

        # Select the next instance from the learner
        ins = next(learner)

        # Check if the instance identifier has not been labeled already
        while ins.identifier in self.env.labeled:
            # This instance has already been labeled my another learner.
            # Skip it and mark as labeled
            learner.set_as_labeled(ins)
            LOGGER.info(
                "The document with key %s was already labeled. Skipping", ins.identifier
            )
            learner = self._choose_learner()
            ins = next(learner)

        # Set the instances as sampled by learner with key al_idx and return the instance
        self._sample_dict[ins.identifier] = self.learners.index(learner)
        return ins

    def set_as_labeled(self, instance: Instance[KT, DT, VT, RT]) -> None:
        self.env.labeled.add(instance)
        self.env.unlabeled.discard(instance)
        if instance.identifier in self._sample_dict:
            learner = self.learners[self._sample_dict[instance.identifier]]
            learner.set_as_labeled(instance)

    def set_as_unlabeled(self, instance: Instance[KT, DT, VT, RT]) -> None:
        self.env.unlabeled.add(instance)
        self.env.labeled.discard(instance)
        if instance.identifier in self._sample_dict:
            learner = self.learners[self._sample_dict[instance.identifier]]
            learner.set_as_unlabeled(instance)
            del self._sample_dict[instance.identifier]

    @classmethod
    def builder(
        cls,
        learner_builders: Sequence[
            Callable[..., ActiveLearner[IT, KT, DT, VT, RT, LT]]
        ],
        probabilities: Sequence[float],
        *_: Any,
        **__: Any,
    ) -> Callable[..., Self]:
        def wrap_func(
            env: AbstractEnvironment[IT, KT, DT, VT, RT, LT], *args, **kwargs
        ):

            learners = [
                builder(env.from_environment(env), *args, **kwargs)
                for builder in learner_builders
            ]
            return cls(env, learners, probabilities)

        return wrap_func


class StrategyEnsemble(
    AbstractEnsemble[IT, KT, DT, VT, RT, LT],
    MLBased[IT, KT, DT, VT, RT, LT, LVT, PVT],
    Generic[IT, KT, DT, VT, RT, LT, LVT, PVT],
):
    _name = "StrategyEnsemble"

    def __init__(
        self,
        env: AbstractEnvironment[IT, KT, DT, VT, RT, LT],
        classifier: AbstractClassifier[KT, VT, LT, LVT, PVT],
        learners: Sequence[
            ActiveLearner[IT, KT, DT, VT, RT, LT],
        ],
        fallback: ActiveLearner[IT, KT, DT, VT, RT, LT],
        probabilities: Sequence[float],
        rng: Any = None,
        *_,
        identifier: Optional[str] = None,
        **__,
    ) -> None:
        super().__init__(env, classifier, fallback)
        self.learners = learners
        self.probabilities = probabilities
        self._rng: Any = get_random_generator(rng)
        self.sampled: Set[KT] = set()
        self._has_ordering: bool = False

    def update_ordering(self) -> bool:
        successful = super().update_ordering()
        if successful:
            self.sampled = set()
        return successful

    def _choose_learner(self) -> ActiveLearner[IT, KT, DT, VT, RT, LT]:
        """Internal functions that selects the next active learner for the next query

        Returns:
            ActiveLearner[IT, KT, DT, VT, RT, LT]: One of the learners from the ensemble
        """
        idxs = np.arange(len(self.learners))
        al_idx: int = self._rng.choice(idxs, size=1, p=self.probabilities)[0]
        learner = self.learners[al_idx]
        return learner

    def __next__(self) -> IT:
        result = super().__next__()
        while result.identifier in self.sampled:
            result = super().__next__()
        self.sampled.add(result.identifier)
        return result


class ILStrategyEnsemble(
    AbstractEnsemble[IT, KT, DT, VT, RT, LT],
    ILMLBased[IT, KT, DT, VT, RT, LT, LMT, PMT],
    Generic[IT, KT, DT, VT, RT, LT, LMT, PMT],
):
    def __init__(
        self,
        env: AbstractEnvironment[IT, KT, DT, VT, RT, LT],
        classifier: il.AbstractClassifier[IT, KT, DT, VT, RT, LT, LMT, PMT],
        learners: Sequence[
            ActiveLearner[IT, KT, DT, VT, RT, LT],
        ],
        fallback: ActiveLearner[IT, KT, DT, VT, RT, LT],
        probabilities: Sequence[float],
        rng: Any = None,
        *_,
        identifier: Optional[str] = None,
        **__,
    ) -> None:
        super().__init__(env, classifier, fallback, identifier=identifier)
        self.learners = learners
        self.probabilities = probabilities
        self._rng = get_random_generator(rng)
        self.sampled: Set[KT] = set()
        self._has_ordering: bool = False

    def update_ordering(self) -> bool:
        successful = super().update_ordering()
        if successful:
            self.sampled = set()
        return successful

    def _choose_learner(self) -> ActiveLearner[IT, KT, DT, VT, RT, LT]:
        """Choose an Active Learning from the options according to
        the given probabilities

        Returns
        -------
        ActiveLearner[IT, KT, DT, VT, RT, LT]
            The chosen ActiveLearner
        """
        idxs = np.arange(len(self.learners))
        al_idx: int = self._rng.choice(idxs, size=1, p=self.probabilities)[0]
        learner = self.learners[al_idx]
        return learner

    def __next__(self) -> IT:
        result = super().__next__()
        while result.identifier in self.sampled:
            result = super().__next__()
        self.sampled.add(result.identifier)
        return result



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------
import os
import sys
sys.path.insert(0, os.path.abspath("..//allib"))
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'ALLib'
copyright = '2021, Michiel Bron MSc'
author = 'Michiel Bron MSc'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosectionlabel",
    "sphinx_autodoc_typehints",
    "sphinx.ext.doctest"
]

autosectionlabel_prefix_document = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('http://docs.scipy.org/doc/numpy', None),
    'pandas': ('https://pandas.pydata.org/docs/', None),
    'sklearn': ('http://scikit-learn.org/stable', (None, './_intersphinx/sklearn-objects.inv')),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = True
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
always_document_param_types = True
typehints_document_rtype = True
autodoc_typehints = "signature"


from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, FrozenSet, Generic, Sequence, Tuple, TypeVar

import numpy as np
from instancelib.instances.base import Instance
from instancelib.labels.base import LabelProvider
from instancelib.labels.memory import MemoryLabelProvider

from ..activelearning import ActiveLearner
from ..activelearning.ml_based import MLBased
from ..utils.func import union

KT = TypeVar("KT")
DT = TypeVar("DT")
VT = TypeVar("VT")
RT = TypeVar("RT")
LT = TypeVar("LT")
LVT = TypeVar("LVT")
PVT = TypeVar("PVT")


class ResultUnit(Enum):
    PERCENTAGE = "Percentage"
    ABSOLUTE = "Absolute"
    FRACTION = "Fraction"


def loss_er(pos_found: int, effort: int, pos_size: int, dataset_size: int) -> float:
    recall_perc = pos_found / pos_size
    inability_loss = (1 - recall_perc) ** 2
    effort_loss = (100 / dataset_size) ** 2 * (effort / (pos_found + 100)) ** 2
    return inability_loss + effort_loss


@dataclass
class BinaryPerformance(Generic[KT]):
    true_positives: FrozenSet[KT]
    true_negatives: FrozenSet[KT]
    false_positives: FrozenSet[KT]
    false_negatives: FrozenSet[KT]

    @property
    def recall(self) -> float:
        tp = len(self.true_positives)
        fn = len(self.false_negatives)
        recall = tp / (tp + fn)
        return recall

    @property
    def precision(self) -> float:
        tp = len(self.true_positives)
        fp = len(self.false_positives)
        try:
            precision = tp / (tp + fp)
        except ZeroDivisionError:
            return 0.0
        return precision

    @property
    def accuracy(self) -> float:
        tp = len(self.true_positives)
        fp = len(self.false_positives)
        fn = len(self.false_negatives)
        tn = len(self.true_negatives)
        accuracy = (tp + tn) / (tp + tn + fp + fn)
        return accuracy

    @property
    def wss(self) -> float:
        tp = len(self.true_positives)
        fp = len(self.false_positives)
        fn = len(self.false_negatives)
        tn = len(self.true_negatives)
        n = tp + fp + fn + tn
        wss = ((tn + fn) / n) - (1 - (tp / (tp + fn)))
        return wss

    @property
    def loss_er(self) -> float:
        recall_perc = self.recall * 100
        R = len(self.true_positives)
        N = len(
            union(
                self.true_positives,
                self.false_positives,
                self.false_negatives,
                self.true_negatives,
            )
        )
        n = len(union(self.true_positives, self.false_positives))
        inability_loss = (100 - recall_perc) ** 2
        effort_loss = (100 / N) ** 2 * (n / (R + 100)) ** 2
        return inability_loss + effort_loss

    @property
    def f1(self) -> float:
        return self.f_beta(beta=1)

    def f_beta(self, beta: int = 1) -> float:
        b2 = beta * beta
        try:
            fbeta = (1 + b2) * (
                (self.precision * self.recall) / ((b2 * self.precision) + self.recall)
            )
        except ZeroDivisionError:
            fbeta = 0.0
        return fbeta


class MultilabelPerformance(Generic[KT, LT]):
    def __init__(self, *label_performances: Tuple[LT, BinaryPerformance[KT]]):
        self.label_dict = {
            label: performance for (label, performance) in label_performances
        }

    @property
    def true_positives(self) -> FrozenSet[KT]:
        keys = union(*(pf.true_positives for pf in self.label_dict.values()))
        return keys

    @property
    def true_negatives(self) -> FrozenSet[KT]:
        keys = union(*(pf.true_negatives for pf in self.label_dict.values()))
        return keys

    @property
    def false_negatives(self) -> FrozenSet[KT]:
        keys = union(*(pf.false_negatives for pf in self.label_dict.values()))
        return keys

    @property
    def false_positives(self) -> FrozenSet[KT]:
        keys = union(*(pf.false_positives for pf in self.label_dict.values()))
        return keys

    @property
    def recall(self) -> float:
        tp = len(self.true_positives)
        fn = len(self.false_negatives)
        recall = tp / (tp + fn)
        return recall

    @property
    def precision(self) -> float:
        tp = len(self.true_positives)
        fp = len(self.false_positives)
        precision = tp / (tp + fp)
        return precision

    @property
    def accuracy(self) -> float:
        tp = len(self.true_positives)
        fp = len(self.false_positives)
        fn = len(self.false_negatives)
        tn = len(self.true_negatives)
        accuracy = (tp + tn) / (tp + tn + fp + fn)
        return accuracy

    @property
    def f1(self) -> float:
        return self.f_beta(beta=1)

    def f_beta(self, beta: int = 1) -> float:
        b2 = beta * beta
        fbeta = (1 + b2) * (
            (self.precision * self.recall) / ((b2 * self.precision) + self.recall)
        )
        return fbeta

    @property
    def f1_macro(self) -> float:
        return self.f_macro(beta=1)

    def f_macro(self, beta=1) -> float:
        average_recall = np.mean([pf.recall for pf in self.label_dict.values()])
        average_precision = np.mean([pf.precision for pf in self.label_dict.values()])
        b2 = beta * beta
        fbeta = (1 + b2) * (
            (average_precision * average_recall)
            / ((b2 * average_precision) + average_recall)
        )
        return fbeta  # type: ignore


def label_metrics(
    truth: LabelProvider[KT, LT],
    prediction: LabelProvider[KT, LT],
    keys: Sequence[KT],
    label: LT,
):
    included_keys = frozenset(keys)
    ground_truth_pos = truth.get_instances_by_label(label).intersection(included_keys)
    pred_pos = prediction.get_instances_by_label(label)
    true_pos = pred_pos.intersection(ground_truth_pos)
    false_pos = pred_pos.difference(true_pos)
    false_neg = ground_truth_pos.difference(true_pos)
    true_neg = included_keys.difference(true_pos, false_pos, false_neg)
    return BinaryPerformance[KT](true_pos, true_neg, false_pos, false_neg)


def classifier_performance(
    learner: MLBased[Any, KT, DT, VT, RT, LT, Any, Any],
    ground_truth: LabelProvider[KT, LT],
    instances: Sequence[Instance[KT, DT, VT, RT]],
) -> Dict[LT, BinaryPerformance[KT]]:
    keys = [ins.identifier for ins in instances]
    labelset = learner.env.labels.labelset
    pred_provider = MemoryLabelProvider[KT, LT](labelset, {}, {})
    predictions = learner.predict(instances)
    for ins, pred in zip(instances, predictions):
        pred_provider.set_labels(ins, *pred)
    performance = {
        label: label_metrics(ground_truth, pred_provider, keys, label)
        for label in labelset
    }
    return performance


def classifier_performance_ml(
    learner: MLBased[Any, KT, DT, VT, RT, LT, Any, Any],
    ground_truth: LabelProvider[KT, LT],
    instances: Sequence[Instance[KT, DT, VT, RT]],
) -> MultilabelPerformance[KT, LT]:
    keys = [ins.identifier for ins in instances]
    labelset = learner.env.labels.labelset
    pred_provider = MemoryLabelProvider[KT, LT](labelset, {}, {})
    predictions = learner.predict(instances)
    for ins, pred in zip(instances, predictions):
        pred_provider.set_labels(ins, *pred)
    performances = [
        (label, label_metrics(ground_truth, pred_provider, keys, label))
        for label in labelset
    ]
    performance = MultilabelPerformance[KT, LT](*performances)
    return performance


def process_performance(
    learner: ActiveLearner[Any, KT, Any, Any, Any, LT], label: LT
) -> BinaryPerformance[KT]:
    labeled = frozenset(learner.env.labeled)
    labeled_positives = frozenset(
        learner.env.get_subset_by_labels(learner.env.labeled, label)
    )
    labeled_negatives = labeled.difference(labeled_positives)

    truth_positives = learner.env.truth.get_instances_by_label(label)

    unlabeled = frozenset(learner.env.unlabeled)
    unlabeled_positives = unlabeled.intersection(
        learner.env.truth.get_instances_by_label(label)
    )
    unlabeled_negatives = unlabeled.difference(unlabeled_positives)

    true_positives = labeled_positives.intersection(truth_positives)
    false_positives = labeled_positives.difference(truth_positives).union(
        labeled_negatives
    )
    false_negatives = truth_positives.difference(labeled_positives)
    true_negatives = unlabeled_negatives

    return BinaryPerformance(
        true_positives, true_negatives, false_positives, false_negatives
    )



from __future__ import annotations

import functools
import itertools
import pickle
import random
import typing as ty
from abc import ABC, abstractmethod
from collections import OrderedDict
from dataclasses import dataclass
from os import PathLike
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Generic,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import instancelib as il
import matplotlib.pyplot as plt  # type: ignore
import numpy as np
import numpy.typing as npt
import pandas as pd
from instancelib.feature_extraction.base import BaseVectorizer
from instancelib.functions.vectorize import vectorize
from instancelib.instances.base import Instance
from tqdm.auto import tqdm

from ..activelearning.base import ActiveLearner
from ..environment.base import AbstractEnvironment
from ..environment.memory import MemoryEnvironment
from ..factory.factory import ObjectFactory
from ..module.component import Component
from ..stopcriterion.base import AbstractStopCriterion
from ..typehints import DT, IT, KT, LT, RT, VT
from .classificationplotter import ClassificationPlotter
from .experiments import ClassificationExperiment, ExperimentIterator
from .initialization import Initializer
from .plotter import AbstractPlotter, ExperimentPlotter


def reset_environment(
    vectorizer: BaseVectorizer[IT],
    environment: AbstractEnvironment[IT, KT, DT, npt.NDArray[Any], RT, LT],
) -> AbstractEnvironment[IT, KT, DT, npt.NDArray[Any], RT, LT]:
    env = MemoryEnvironment.from_environment_only_data(environment)
    vectorize(vectorizer, env, True, 200)  # type: ignore
    return env


def initialize_tar_simulation(
    factory: ObjectFactory,
    al_config: Mapping[str, Any],
    fe_config: Mapping[str, Any],
    initializer: Initializer[IT, KT, LT],
    env: AbstractEnvironment[IT, KT, DT, npt.NDArray[Any], RT, LT],
    pos_label: LT,
    neg_label: LT,
) -> Tuple[
    ActiveLearner[IT, KT, DT, npt.NDArray[Any], RT, LT],
    BaseVectorizer[Instance[KT, DT, npt.NDArray[Any], RT]],
]:
    """Build and initialize an Active Learning method.

    Parameters
    ----------
    factory : ObjectFactory
        The factory method that builds the components
    al_config : Mapping[str, Any]
        The dictionary that declares the configuration of the Active Learning component
    fe_config : Mapping[str, Any]
        The dictionary that declares the configuration of the Feature Extraction component
    initializer : Initializer[IT, KT, LT]
        The function that determines how and which initial knowledge should be supplied to
        the Active Learner
    env : AbstractEnvironment[KT, DT, npt.NDArray[Any], DT, LT]
        The environment on which we should simulate
    pos_label : LT
        The label of the positive class
    neg_label : LT

    Returns
    -------
    Tuple[ActiveLearner[KT, DT, npt.NDArray[Any], DT, LT], BaseVectorizer[Instance[KT, DT, npt.NDArray[Any], DT]]]
        A tuple that contains:

        - An :class:`~allib.activelearning.base.ActiveLearner` object according
            to the configuration in `al_config`
        - An :class:`~allib.feature_extraction.base.BaseVectorizer` object according
            to the configuration in `fe_config`
    """
    # Get the active learner builder and feature extraction models
    learner_builder: Callable[
        ..., ActiveLearner[IT, KT, DT, npt.NDArray[Any], RT, LT]
    ] = factory.create(Component.ACTIVELEARNER, **al_config)
    vectorizer: BaseVectorizer[Instance[KT, DT, npt.NDArray[Any], RT]] = factory.create(
        Component.FEATURE_EXTRACTION, **fe_config
    )

    ## Copy the data to memory
    start_env = reset_environment(vectorizer, env)

    # Build the Active Learner object
    learner = learner_builder(start_env, pos_label=pos_label, neg_label=neg_label)

    # Initialize the learner with initial knowledge
    learner = initializer(learner)
    return learner, vectorizer


class TarSimulator(Generic[IT, KT, DT, VT, RT, LT]):
    plotter: ExperimentPlotter[LT]
    experiment: ExperimentIterator
    output_pkl_path: Optional[Path]
    output_pdf_path: Optional[Path]
    plot_interval: int

    def __init__(
        self,
        experiment: ExperimentIterator[IT, KT, DT, VT, RT, LT],
        plotter: ExperimentPlotter[LT],
        max_it: Optional[int] = None,
        print_enabled=False,
        output_path: Optional[Path] = None,
        output_pdf_path: Optional[Path] = None,
        plot_interval: int = 20,
    ) -> None:
        self.experiment = experiment
        self.plotter = plotter
        self.max_it = max_it
        self.print_enabled = print_enabled
        self.output_pkl_path = output_path
        self.output_pdf_path = output_pdf_path
        self.plot_interval = plot_interval

    @property
    def _debug_finished(self) -> bool:
        if self.max_it is None:
            return False
        return self.experiment.it > self.max_it

    def simulate(self) -> None:
        with tqdm(total=len(self.experiment.learner.env.dataset)) as pbar:
            pbar.update(self.experiment.learner.len_labeled)
            while not self.experiment.finished and not self._debug_finished:
                result = self.experiment()
                self.plotter.update(self.experiment, result)
                if self.print_enabled:
                    self.plotter.print_last_stats()
                pbar.update(1)
                if self.output_pkl_path is not None:
                    with self.output_pkl_path.open("wb") as fh:
                        pickle.dump(self.plotter, fh)
                if (
                    self.experiment.it % self.plot_interval == 0
                    and self.output_pdf_path is not None
                ):
                    self.plotter.show(filename=self.output_pdf_path)


class ClassificationSimulator(Generic[IT, KT, DT, VT, RT, LT]):
    plotter: ClassificationPlotter[LT]
    experiment: ClassificationExperiment[IT, KT, DT, VT, RT, LT]

    def __init__(
        self,
        experiment: ClassificationExperiment[IT, KT, DT, VT, RT, LT],
        plotter: ClassificationPlotter[LT],
        max_it: Optional[int] = None,
        print_enabled=False,
    ) -> None:
        self.experiment = experiment
        self.plotter = plotter
        self.max_it = max_it
        self.print_enabled = print_enabled

    @property
    def _debug_finished(self) -> bool:
        if self.max_it is None:
            return False
        return self.experiment.it > self.max_it

    def simulate(self) -> None:
        first_learner = next(iter(self.experiment.learners.values()))
        with tqdm(total=len(first_learner.env.dataset)) as pbar:
            pbar.update(first_learner.len_labeled)
            while not self.experiment.finished and not self._debug_finished:
                result = self.experiment()
                self.plotter.update(self.experiment, result)
                if self.print_enabled:
                    self.plotter.print_last_stats()
                pbar.update(1)


def multilabel_all_non_empty(
    learner: ActiveLearner[Any, Any, Any, Any, Any, Any], count: int
) -> bool:
    provider = learner.env.labels
    non_empty = all(
        [provider.document_count(label) > count for label in provider.labelset]
    )
    return non_empty



from typing import Any, Dict
from ..module import ModuleCatalog as Cat

autotar = {
    "paradigm": Cat.AL.Paradigm.CUSTOM,
    "method": Cat.AL.CustomMethods.AUTOTAR,
    "k_sample": 100,
    "batch_size": 20, 
}

autostop = {
    "paradigm": Cat.AL.Paradigm.CUSTOM,
    "method": Cat.AL.CustomMethods.AUTOSTOP,
    "k_sample": 100,
    "batch_size": 1, 
}